# Copyright (C) 2017 Cancer Care Associates

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

from pymedphys._imports import numpy as np


def find_scan_index(file_contents):
    """Searches through the mephysto file for where the BEGIN_SCAN and END_SCAN
    appear. This region contains all of the information for a given scan within
    the mephysto file. Returns a list of ranges which span each scan region.
    """
    # Find the positions of all BEGIN_SCAN
    # Demo of match -- https://regex101.com/r/lR4pS2/2
    # Demo of ignoring "BEGIN_SCAN_DATA" -- https://regex101.com/r/lR4pS2/4
    # Basic guide to regular expressions (regex):
    #    http://www.regular-expressions.info/quickstart.html
    begin_scan_index = np.array(
        [
            i
            for i, item in enumerate(file_contents)
            if re.search(r"^\tBEGIN_SCAN\s\s\d+$", item)
        ]
    ).astype(int)

    # Find the positions of all END_SCAN
    # Regex demo here -- https://regex101.com/r/lR4pS2/3
    end_scan_index = np.array(
        [
            i
            for i, item in enumerate(file_contents)
            if re.search(r"^\tEND_SCAN\s\s\d+$", item)
        ]
    ).astype(int)

    # Convert the indices into the range type allowing for easy looping
    scan_index = [
        range(begin_scan_index[i] + 1, end_scan_index[i])
        for i in range(len(begin_scan_index))
    ]

    return scan_index


def find_data_index(file_contents):
    """Searches through the mephysto file for where the BEGIN_DATA and END_DATA
    appear. This region contains only the raw data in column format. Returns a
    list of ranges which span each scan region.
    """
    # Find the positions of BEGIN_DATA
    # Demo of match -- https://regex101.com/r/lR4pS2/5
    begin_data_index = np.array(
        [
            i
            for i, item in enumerate(file_contents)
            if re.search(r"^\t\tBEGIN_DATA$", item)
        ]
    ).astype(int)

    # Find the positions of END_DATA
    # Demo of match -- https://regex101.com/r/lR4pS2/6
    end_data_index = np.array(
        [
            i
            for i, item in enumerate(file_contents)
            if re.search(r"^\t\tEND_DATA$", item)
        ]
    ).astype(int)

    # Convert the indices into the range type allowing for easy looping
    data_index = [
        range(begin_data_index[i] + 1, end_data_index[i])
        for i in range(len(begin_data_index))
    ]

    return data_index


def pull_mephysto_item(string, file_contents):
    """Steps through each scan region searching for a mephysto parameter that
    matches the requested string. Returns an array filled with the results for
    all scans that have a match. If a scan does not have a parameter matching
    the request then np.nan is returned.
    """
    scan_index = find_scan_index(file_contents)

    # Format the input string for use in a regex search
    string_test = re.escape(string)

    result = []
    # Loop over each scan index searching for requested parameter
    for index in scan_index:
        # See if any parameters match within this scan
        match = np.array(
            [
                re.match(r"^\t\t" + string_test + "=(.*)$", item) is not None
                for item in file_contents[index]
            ]
        )

        # If there is a single match then return the result
        if np.sum(match) == 1:
            relevant_line = file_contents[index][match][0]
            # Return the value that is after the equal sign
            # Example -- https://regex101.com/r/lR4pS2/7
            result.append(
                re.search(r"^\t\t" + string_test + "=(.*)$", relevant_line).group(1)
            )

        # If there is no matches return np.nan
        elif np.sum(match) == 0:
            result.append(np.nan)

        # If there is more than one match raise an error
        else:
            raise ValueError("More than one item has this label")

    return np.array(result)


def pull_mephysto_number(string, file_contents):
    """Pulls data using pull_mephysto_item and returns the result to the user
    as a float.
    """
    result = pull_mephysto_item(string, file_contents)
    return result.astype(float)


def pull_mephysto_data(file_contents):
    """Pull the distance and relative dose from the mephysto file contents.
    Will only be able to return data if the data is stored in two columns. Some
    mephysto scans use three or more columns, for these this function will have
    to be updated.
    """
    scan_index = find_scan_index(file_contents)
    data_index = find_data_index(file_contents)

    # Confirm that each data index is within the expected scan index. Raise an
    # error if it is not.
    for i in range(len(scan_index)):  # pylint: disable = consider-using-enumerate
        assert (scan_index[i][0] < data_index[i][0]) & (
            scan_index[i][-1] > data_index[i][-1]
        )

    distance = []
    # Loop through the data index and return all of the first column
    for index in data_index:
        # Demo of regex -- https://regex101.com/r/lR4pS2/10
        distance.append(
            np.array(
                [
                    re.search(
                        r"^\t\t\t([-+]?\d+\.\d+)\t\t(\d+\.\d+E[-+]?\d+)", item
                    ).group(1)
                    for item in file_contents[index]
                ]
            ).astype(float)
        )

    relative_dose = []
    # Loop through the data index and return all of the second column
    for index in data_index:
        relative_dose.append(
            np.array(
                [
                    re.search(
                        r"^\t\t\t([-+]?\d+\.\d+)\t\t(\d+\.\d+E[-+]?\d+)", item
                    ).group(2)
                    for item in file_contents[index]
                ]
            ).astype(float)
        )

    return distance, relative_dose



from pymedphys._imports import numpy as np


def running_mean(x, N):
    out = np.zeros_like(x, dtype=np.float64)
    dim_len = x.shape[0]
    for i in range(dim_len):
        if N % 2 == 0:
            a, b = i - (N - 1) // 2, i + (N - 1) // 2 + 2
        else:
            a, b = i - (N - 1) // 2, i + (N - 1) // 2 + 1

        # cap indices to min and max indices
        a = max(0, a)
        b = min(dim_len, b)
        out[i] = np.mean(x[a:b])
    return out



# Copyright (C) 2019 Cancer Care Associates

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os

from pymedphys._trf.manage import already_indexed_path


def test_file_logic():
    diagnostics = "/a/path/to/diagnostics"
    to_be_indexed = os.path.join(diagnostics, "to_be_indexed")
    already_indexed = os.path.join(diagnostics, "already_indexed")

    machine_zip_file_path = "machine/archive.zip"
    current_location = os.path.join(to_be_indexed, machine_zip_file_path)
    expected_new = os.path.abspath(os.path.join(already_indexed, machine_zip_file_path))

    converted = already_indexed_path(current_location, to_be_indexed, already_indexed)

    assert converted == expected_new



# Copyright (C) 2019 Cancer Care Associates

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from copy import deepcopy

from .core import (
    get_fraction_group_beam_sequence_and_meterset,
    get_fraction_group_index,
)


def convert_to_one_fraction_group(dicom_dataset, fraction_group_number):
    created_dicom = deepcopy(dicom_dataset)

    beam_sequence, _ = get_fraction_group_beam_sequence_and_meterset(
        dicom_dataset, fraction_group_number
    )

    created_dicom.BeamSequence = beam_sequence

    fraction_group_index = get_fraction_group_index(
        dicom_dataset, fraction_group_number
    )

    fraction_group = created_dicom.FractionGroupSequence[fraction_group_index]
    created_dicom.FractionGroupSequence = [fraction_group]

    return created_dicom



# Copyright (C) 2019 Cancer Care Associates

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import numpy as np

from pymedphys._mocks.profiles import (
    create_profile_function,
    create_square_field_function,
    gaussian_cdf,
    scaled_penumbra_sig,
)


def test_scaled_penumbra():
    profile_shoulder_edges = [0.7, 0.8, 0.9]

    for profile_shoulder_edge in profile_shoulder_edges:
        sig = scaled_penumbra_sig(profile_shoulder_edge)
        np.testing.assert_allclose(
            gaussian_cdf([-0.5, 0, 0.5], sig=sig),
            [1 - profile_shoulder_edge, 0.5, profile_shoulder_edge],
        )


def test_profile_function():
    x = [-5, -1, 0, 1, 3, 5, 7, 9, 10, 11, 15]
    centre = 5
    field_width = 10
    penumbra_width = 2
    expected_profile_values = [0, 0.2, 0.5, 0.8, 1, 1, 1, 0.8, 0.5, 0.2, 0]

    profile = create_profile_function(centre, field_width, penumbra_width)

    np.testing.assert_allclose(expected_profile_values, profile(x), atol=0.01)


def test_field_function():
    centre = [5, 5]
    side_length = 2
    penumbra_width = 0.3
    rotation = 10

    field = create_square_field_function(centre, side_length, penumbra_width, rotation)

    xx, yy = np.meshgrid(np.linspace(3, 7, 10), [4, 5, 6])

    expected_results = [
        [0.0, 0.0, 0.11, 0.29, 0.45, 0.62, 0.64, 0.06, 0.0, 0.0],
        [0.0, 0.0, 0.3, 0.97, 1.0, 1.0, 0.97, 0.3, 0.0, 0.0],
        [0.0, 0.0, 0.06, 0.64, 0.62, 0.45, 0.29, 0.11, 0.0, 0.0],
    ]

    field_values = np.round(field(xx, yy), 2)

    assert np.all(field_values == expected_results)



# Copyright (C) 2015-2018 Simon Biggs
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from pymedphys._imports import numpy as np


def calculate_coordinates_shell(distance, num_dimensions, distance_step_size):
    """Create the shell of coordinate shifts for the given testing distance.

    Coordinate shifts are determined to check the evaluation dose for a
    given distance, dimension, and step size
    """
    if num_dimensions == 1:
        return calculate_coordinates_shell_1d(distance)

    if num_dimensions == 2:
        return calculate_coordinates_shell_2d(distance, distance_step_size)

    if num_dimensions == 3:
        return calculate_coordinates_shell_3d(distance, distance_step_size)

    raise ValueError("No valid dimension")


def calculate_coordinates_shell_1d(distance):
    """Output the two points that are of the defined distance in one-dimension
    """
    if distance == 0:
        x_coords = np.array([0])
    else:
        x_coords = np.array([distance, -distance])

    return (x_coords,)


def calculate_coordinates_shell_2d(distance, distance_step_size):
    """Create points along the circumference of a circle. The spacing
    between points is not larger than the defined distance_step_size
    """
    amount_to_check = np.ceil(2 * np.pi * distance / distance_step_size).astype(int) + 1
    theta = np.linspace(0, 2 * np.pi, amount_to_check + 1)[:-1:]
    x_coords = distance * np.cos(theta)
    y_coords = distance * np.sin(theta)

    return (x_coords, y_coords)


def calculate_coordinates_shell_3d(distance, distance_step_size):
    """Create points along the surface of a sphere (a shell) where no gap
    between points is larger than the defined distance_step_size"""

    number_of_rows = np.ceil(np.pi * distance / distance_step_size).astype(int) + 1

    elevation = np.linspace(0, np.pi, number_of_rows)
    row_radii = distance * np.sin(elevation)
    row_circumference = 2 * np.pi * row_radii
    amount_in_row = np.ceil(row_circumference / distance_step_size).astype(int) + 1

    x_coords = []
    y_coords = []
    z_coords = []
    for i, phi in enumerate(elevation):
        azimuth = np.linspace(0, 2 * np.pi, amount_in_row[i] + 1)[:-1:]
        x_coords.append(distance * np.sin(phi) * np.cos(azimuth))
        y_coords.append(distance * np.sin(phi) * np.sin(azimuth))
        z_coords.append(distance * np.cos(phi) * np.ones_like(azimuth))

    return (np.hstack(x_coords), np.hstack(y_coords), np.hstack(z_coords))



# Copyright (C) 2019 Simon Biggs

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

import numpy as np

from pymedphys._metersetmap.metersetmap import calc_metersetmap

MU = [0, 30]

MLC = np.array([[[4, -3], [4, -3], [4, -3]], [[-3, 4], [-3, 4], [-3, 4]]])

JAW = np.array([[3, 3], [3, 3]])

LEAF_PAIR_WIDTHS = [2, 2, 2]


def test_max_leaf_gap():
    max_leaf_gap_init = 2 * np.max(np.abs(MLC))
    grid_resolution = 2
    how_many_to_test = 20

    max_leaf_gap_init = (
        np.ceil(max_leaf_gap_init / 2 / grid_resolution) * 2 * grid_resolution
    )

    multiple_max_leaf_gaps = np.arange(
        max_leaf_gap_init,
        max_leaf_gap_init + how_many_to_test * 2 * grid_resolution,
        2 * grid_resolution,
    )

    init_metersetmap = calc_metersetmap(
        MU,
        MLC,
        JAW,
        leaf_pair_widths=LEAF_PAIR_WIDTHS,
        max_leaf_gap=max_leaf_gap_init,
        grid_resolution=grid_resolution,
    )

    for i, max_leaf_gap in enumerate(multiple_max_leaf_gaps):
        metersetmap = calc_metersetmap(
            MU,
            MLC,
            JAW,
            leaf_pair_widths=LEAF_PAIR_WIDTHS,
            max_leaf_gap=max_leaf_gap,
            grid_resolution=grid_resolution,
        )

        assert not np.all(metersetmap == 0)
        if i != 0:
            assert np.all(metersetmap[:, i:-i] == init_metersetmap)

    with pytest.raises(ValueError):
        calc_metersetmap(
            MU,
            MLC,
            JAW,
            leaf_pair_widths=LEAF_PAIR_WIDTHS,
            max_leaf_gap=max_leaf_gap_init + 1,
            grid_resolution=grid_resolution,
        )



from .core import (
    BASELINE_DICOM_REPEATERS_DICT_FILEPATH,
    DICOM_SOP_CLASS_NAMES_MODE_PREFIXES,
    NotInBaselineError,
    get_baseline_dicom_dict,
    get_baseline_dicom_repeaters_dict,
    get_baseline_dict_entry,
    get_baseline_keyword_vr_dict,
)
from .orientation import IMAGE_ORIENTATION_MAP
from .uuid import PYMEDPHYS_ROOT_UID



import logging
import pathlib
import re
import socket
import time
import traceback

from . import patients

BUFFER_SIZE = 256
ICOM_PORT = 1706


def save_an_icom_batch(date_pattern, ip_directory, data_to_save):
    if not date_pattern.match(data_to_save[8:26]):
        raise ValueError("Unexpected iCOM stream format")

    counter = str(int(data_to_save[26])).zfill(3)
    filepath = ip_directory.joinpath(f"{counter}.txt")

    with open(filepath, "bw+") as f:
        f.write(data_to_save)

    logging.debug("Saved stream to %(filepath)s", {"filepath": filepath})


def get_start_location_from_date_span(span):
    return span[0] - 8


def initialise_socket(ip):
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.connect((ip, ICOM_PORT))
    s.settimeout(10)
    logging.info(s)

    return s


def listen(ip, data_dir):
    date_pattern = re.compile(rb"\d\d\d\d-\d\d-\d\d\d\d:\d\d:\d\d")

    data_dir = pathlib.Path(data_dir)
    live_dir = data_dir.joinpath("live")
    patients_dir = data_dir.joinpath("patients")

    patient_icom_data = patients.PatientIcomData(patients_dir)

    def archive_by_patient(ip, data):
        patient_icom_data.update_data(ip, data)

    ip_directory = live_dir.joinpath(ip)
    ip_directory.mkdir(exist_ok=True, parents=True)

    s = initialise_socket(ip)

    try:
        data = b""

        while True:
            try:
                data += s.recv(BUFFER_SIZE)
            except socket.timeout:
                logging.warning("Socket connection timed out, retrying connection")
                logging.info(s)
                s.close()
                logging.info(s)
                s = initialise_socket(ip)
                continue

            matches = date_pattern.finditer(data)
            try:
                span = next(matches).span()
            except StopIteration:
                continue

            previous_start_location = get_start_location_from_date_span(span)
            for match in matches:
                new_start_location = get_start_location_from_date_span(match.span())
                data_to_save = data[previous_start_location:new_start_location]

                save_an_icom_batch(date_pattern, ip_directory, data_to_save)
                archive_by_patient(ip, data_to_save)

                previous_start_location = new_start_location

            data = data[previous_start_location::]

    finally:
        s.close()
        logging.info(s)


def listen_cli(args):
    while True:
        try:
            listen(args.ip, args.directory)
        except KeyboardInterrupt:
            raise
        except:  # pylint: disable = bare-except
            traceback.print_exc()

        logging.warning(
            "The iCOM listener dropped out. Will wait 15 minutes, and "
            "then retry connection."
        )

        time.sleep(60 * 15)



import logging
import lzma
import pathlib
import traceback

import pymedphys

from . import extract, observer

# TODO: Convert logging to use lazy formatting
# see https://docs.python.org/3/howto/logging.html#optimization


class NoMUDelivered(ValueError):
    pass


class UnableToReadIcom(ValueError):
    pass


def validate_data(data_to_be_saved):
    logging.debug("Validating an iCOM dataset.")

    try:
        delivery = pymedphys.Delivery.from_icom(data_to_be_saved)
        logging.debug("iCOM dataset was found to be valid.")
    except Exception as _:
        logging.debug("Was not able to transform the iCOM dataset.")

        traceback.print_exc()
        raise UnableToReadIcom()

    if len(delivery.mu) == 0:
        logging.debug("Was able to read the iCOM dataset, but no MU was delivered.")

        raise NoMUDelivered()

    return delivery


def save_patient_data(start_timestamp, patient_data, output_dir: pathlib.Path):
    _, patient_id = extract.extract(patient_data[0], "Patient ID")

    logging.debug(
        "When preparing patient record to be saved, the patient id was "
        "%(patient_id)s",
        {"patient_id": patient_id},
    )

    for data in patient_data:
        _, patient_name = extract.extract(data, "Patient Name")
        if not patient_name is None:
            break

    logging.debug(
        "When preparing patient record to be saved, the patient name was "
        "%(patient_name)s",
        {"patient_name": patient_name},
    )

    patient_dir = output_dir.joinpath(f"{patient_id}_{patient_name}")
    patient_dir.mkdir(parents=True, exist_ok=True)

    logging.debug(
        "The patient archive directory to be used is %(patient_dir)s.",
        {"patient_dir": patient_dir},
    )

    reformatted_timestamp = (
        start_timestamp.replace(":", "").replace("T", "_").replace("-", "")
    )
    filename = patient_dir.joinpath(f"{reformatted_timestamp}.xz")

    data = b""
    for item in patient_data:
        data += item

    try:
        delivery = validate_data(data)
        logging.info(  # pylint: disable = logging-fstring-interpolation
            f"Delivery with a total MU of {round(delivery.mu[-1],1)} for "
            f"{patient_name} ({patient_id}) is being saved within "
            f"{filename}."
        )
    except NoMUDelivered as _:
        logging.info(  # pylint: disable = logging-fstring-interpolation
            "No MU delivered, not saving delivery data for "
            f"{patient_name} ({patient_id})."
        )

        return

    except UnableToReadIcom as _:
        new_location = filename.parent.parent.joinpath(
            "unknown_error_in_record", filename.parent.name, filename.name
        )
        new_dir = new_location.parent
        new_dir.mkdir(parents=True, exist_ok=True)

        filename = new_location

        logging.warning(  # pylint: disable = logging-fstring-interpolation
            "Unknown error within the record for "
            f"{patient_name} ({patient_id}). "
            f"Will instead save the record within {str(filename)}."
        )

    with lzma.open(filename, "w") as f:
        f.write(data)


class PatientIcomData:
    def __init__(self, output_dir):
        self._data = {}
        self._usage_start = {}
        self._current_patient_data = {}
        self._output_dir = pathlib.Path(output_dir)

    def update_data(self, ip, data):
        try:
            if self._data[ip][-1][26] == data[26]:
                logging.warning("Skip this data item, duplicate of previous data item.")
                if self._data[ip][-1] != data:
                    raise ValueError("Duplicate ID, but not duplicate data!")

                return

            if (self._data[ip][-1][26] + 1) % 256 != data[26]:
                raise ValueError("Data stream appears to be arriving out of order")
            self._data[ip].append(data)
        except KeyError:
            self._data[ip] = [data]

        timestamp = data[8:26].decode()
        shrunk_data, patient_id = extract.extract(data, "Patient ID")
        shrunk_data, patient_name = extract.extract(shrunk_data, "Patient Name")
        shrunk_data, machine_id = extract.extract(shrunk_data, "Machine ID")
        logging.info(  # pylint: disable = logging-fstring-interpolation
            f"IP: {ip} | Timestamp: {timestamp} | "
            f"Patient ID: {patient_id} | "
            f"Patient Name: {patient_name} | Machine ID: {machine_id}"
        )

        try:
            usage_start = self._usage_start[ip]
        except KeyError:
            usage_start = None

        if patient_id is not None:
            if usage_start is None:
                self._current_patient_data[ip] = []

                timestamp = data[8:26].decode()
                iso_timestamp = f"{timestamp[0:10]}T{timestamp[10::]}"
                self._usage_start[ip] = iso_timestamp

                logging.debug(
                    "Starting data collection for patient id %(patient_id)s. "
                    "Recording started at %(usage_start)s.",
                    {"usage_start": self._usage_start[ip], "patient_id": patient_id},
                )

            self._current_patient_data[ip].append(data)

            logging.debug(
                "iCOM stream appended to the data being collected for "
                "patient id %(patient_id)s.",
                {"patient_id": patient_id},
            )

        elif not usage_start is None:
            logging.debug(
                "Delivery that started at %(usage_start)s appears to "
                "have completed.",
                {"usage_start": usage_start},
            )

            save_patient_data(
                usage_start, self._current_patient_data[ip], self._output_dir
            )
            self._current_patient_data[ip] = None
            self._usage_start[ip] = None

        else:
            logging.debug("No delivery is currently being recorded.")


def archive_by_patient(directories_to_watch, output_dir):
    patient_icom_data = PatientIcomData(output_dir)

    def archive_by_patient_callback(ip, data):
        patient_icom_data.update_data(ip, data)

    observer.observe_with_callback(directories_to_watch, archive_by_patient_callback)



"""Streamlit Gui for Pseudonymise
"""
# Copyright (C) 2020 Stuart Swerdloff

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import base64
import datetime
import io
import pathlib

from zipfile import ZIP_DEFLATED, ZipFile

from pymedphys._imports import pydicom
from pymedphys._imports import streamlit as st

from pymedphys._dicom.anonymise import anonymise_dataset
from pymedphys._dicom.constants.core import DICOM_SOP_CLASS_NAMES_MODE_PREFIXES
from pymedphys._dicom.utilities import remove_file
from pymedphys.experimental import pseudonymisation as pseudonymisation_api


def link_to_zipbuffer_download(filename: str, zip_bytes: bytes):
    """Generates a link allowing the bytes to be downloaded using the specified filename
    Assumes that the data is in zip format

    Parameters
    ----------
    filename : str
        default filename to use
    zip_bytes : bytes
        the bytes from an IOBytes in which a ZipFile wrote its contents, or the bytes
        read directly from a zip file on disk
    """
    b64 = base64.b64encode(zip_bytes).decode()
    href = f"<a href=\"data:file/zip;base64,{b64}\" download='{filename}'>\
        Click to download {filename}\
    </a>"
    st.sidebar.markdown(href, unsafe_allow_html=True)


def build_pseudonymised_file_name(ds_input: pydicom.dataset.Dataset):
    """Create the base filename for the pseudonymised file

    Parameters
    ----------
    ds_input : pydicom.dataset.Dataset
        The non-pseudonymised DICOM data

    Returns
    -------
    anon_filename : str
        name for the pseudonymised file, which can be used
        for addition to a zip
    """
    pseudo_sop_instance_uid = pseudonymisation_api.pseudonymisation_dispatch[
        "UI"
    ](  # type: ignore
        ds_input.SOPInstanceUID
    )

    sop_class_uid: pydicom.dataelem.DataElement = ds_input.SOPClassUID

    mode_prefix = DICOM_SOP_CLASS_NAMES_MODE_PREFIXES[sop_class_uid.name]
    anon_filename = f"{mode_prefix}.{pseudo_sop_instance_uid}_Anonymised.dcm"

    return anon_filename


def _zip_pseudo_fifty_mbytes(file_buffer_list: list, zip_bytes_io: io.BytesIO):
    """Pseudonymises the contents of the file_buffer_list (list of DICOM files)
    and places the pseudonymised files in to a zip.

    Parameters
    ----------
    file_buffer_list : list
        List of DICOM file buffers from streamlit file_uploader to pseudonymise
    zip_bytes_io : io.BytesIO
        An in memory file like object to be used for storing the Zip

    """

    bad_data = False
    file_count = 0
    keywords = pseudonymisation_api.get_default_pseudonymisation_keywords()
    keywords.remove("PatientSex")
    strategy = pseudonymisation_api.pseudonymisation_dispatch
    zip_stream = zip_bytes_io

    with ZipFile(zip_stream, mode="w", compression=ZIP_DEFLATED) as myzip:
        for uploaded_file_buffer in file_buffer_list:
            file_count += 1

            # don't close the buffers.  Streamlit provides the user with control over that.
            # might be appropriate to close the buffers in some circumstances,
            # but then when the user goes to close the buffer (click x on screen)
            # there will be an error.

            try:
                original_file_name = uploaded_file_buffer.name
                ds_input: pydicom.FileDataset = pydicom.dcmread(
                    uploaded_file_buffer, force=True
                )

                anonymise_dataset(
                    ds_input,
                    delete_private_tags=True,
                    delete_unknown_tags=True,
                    copy_dataset=False,  # do the work in place.  less memory used and we're disposing shortly anyway
                    identifying_keywords=keywords,
                    replacement_strategy=strategy,
                )
                temp_anon_filepath = build_pseudonymised_file_name(ds_input)
                in_memory_temp_file = io.BytesIO()
                anon_filename = pathlib.Path(temp_anon_filepath).name
                pydicom.dcmwrite(in_memory_temp_file, ds_input)
            except (KeyError, IOError, ValueError) as e_info:
                print(e_info)
                print(f"While processing {original_file_name}")
                bad_data = True
                break
            myzip.writestr(
                anon_filename,
                in_memory_temp_file.getvalue(),
                compress_type=ZIP_DEFLATED,
            )
            in_memory_temp_file.close()
    return bad_data


def pseudonymise_buffer_list(file_buffer_list: list):
    """Pseudonymises the contents of the file_buffer_list (list of DICOM files)
    and places the pseudonymised files in to a set of zips, each less than 50 MBytes.
    Those zips are then made available for download through a set of links in the
    streamlit.sidebar

    The 50MByte limit is imposed by the href download link limit.

    The compression used is DEFLATE, but the uncompressed contents are kept at just under 50 MBytes
    If a single original uncompressed file is > 50MByte, and does not compress to under 50 MByte
    that file will fail to be made available for download, and may cause the entire pseudonymisation
    attempt to fail.

    Parameters
    ----------
    file_buffer_list : list
        DICOM files that were uploaded using streamlit.file_uploader
    """

    if file_buffer_list is not None and len(file_buffer_list) > 0:
        my_date_time = datetime.datetime.now()
        str_now_datetime = my_date_time.strftime("%Y%m%d_%H%M%S")
        zipfile_basename = f"Pseudonymised_{str_now_datetime}"

        bad_data = False
        index_to_fifty_mbyte_increment = _gen_index_list_to_fifty_mbyte_increment(
            file_buffer_list
        )

        st.write(index_to_fifty_mbyte_increment)

        zip_count = 0
        start_index = 0
        for end_index in index_to_fifty_mbyte_increment:
            if start_index == end_index:
                break
            zip_count += 1
            zipfile_name = f"{zipfile_basename}.{zip_count}.zip"
            zip_bytes_io = io.BytesIO()
            bad_data = _zip_pseudo_fifty_mbytes(
                file_buffer_list[start_index:end_index], zip_bytes_io
            )
            start_index = end_index
            if bad_data:
                if zip_bytes_io is not None:
                    zip_bytes_io.close()
                    del zip_bytes_io
                else:
                    remove_file(zipfile_name)
                st.text("Problem processing DICOM data")
            else:
                if zip_bytes_io is not None:
                    link_to_zipbuffer_download(zipfile_name, zip_bytes_io.getvalue())
                    zip_bytes_io.close()
                    del zip_bytes_io


def _gen_index_list_to_fifty_mbyte_increment(file_buffer_list):
    """Create a list of indexes marking the ~50 MByte increments of cumulative size
    of the file_buffers provided by the streamlit.file_uploader

    Parameters
    ----------
    file_buffer_list : ``list``
        list of file_buffers (derived from io.BytesIO) obtained from streamlit.file_uploader

    Returns
    -------
    ``list``
        list of indexes marking the ~50 MByte increments, including the last index
    """

    file_count = 0
    index_to_fifty_mbyte_increment = list()
    size = 0
    for uploaded_file_buffer in file_buffer_list:
        file_count += 1
        size += uploaded_file_buffer.size
        if size > 50000000:
            index_to_fifty_mbyte_increment.append(file_count)
            size = 0

    if size != 0:
        index_to_fifty_mbyte_increment.append(file_count)

    return index_to_fifty_mbyte_increment


def main():
    st.set_option("deprecation.showfileUploaderEncoding", False)

    uploaded_file_buffer_list = st.file_uploader(
        "Files to pseudonymise, refresh page after downloading zip(s)",
        ["dcm"],
        accept_multiple_files=True,
    )

    if st.button("Pseudonymise", key="PseudonymiseButton"):
        pseudonymise_buffer_list(uploaded_file_buffer_list)
        uploaded_file_buffer_list.clear()

    # this deletion never worked.  which motivated switch to using IOBytes for ZipFile
    # if st.button(f"Delete Zip(s)", key="DeleteZip"):
    #     for zip_path_str in zip_path_list:
    #         remove_file(zip_path_str)


if __name__ == "__main__":
    main()



import functools
import getpass
import json
import pathlib

from pymedphys._imports import keyring, requests

ZENODO_HOSTNAME = "zenodo.org"
BASE_URL = f"https://{ZENODO_HOSTNAME}/api/records/"
HERE = pathlib.Path(__file__).resolve().parent


@functools.lru_cache()
def get_zenodo_file_urls(record_name):
    files = get_all_zenodo_file_data(record_name)
    urls = {record["key"]: record["links"]["self"] for record in files}

    return urls


@functools.lru_cache()
def get_all_zenodo_file_data(record_name):
    record_id = get_zenodo_record_id(record_name)
    files = get_all_zenodo_file_data_by_id(record_id)

    return files


@functools.lru_cache()
def get_all_zenodo_file_data_by_id(record_id):
    response = requests.get(f"{BASE_URL}{record_id}")
    data = json.loads(response.text)
    files = data["files"]

    return files


@functools.lru_cache()
def get_zenodo_file_md5s(record_name):
    files = get_all_zenodo_file_data(record_name)
    checksums = {
        record["key"]: record["checksum"].replace("md5:", "") for record in files
    }

    return checksums


@functools.lru_cache()
def get_zenodo_record_id(record_name):
    with open(HERE.joinpath("zenodo.json"), "r") as zenodo_file:
        zenodo = json.load(zenodo_file)

    try:
        record_id = zenodo[record_name]
    except KeyError:
        raise ValueError(
            "This Zenodo record isn't recorded within this version of PyMedPhys."
        )

    return record_id


def update_zenodo_record_id(record_name, record_id):
    with open(HERE.joinpath("zenodo.json"), "r") as zenodo_file:
        zenodo = json.load(zenodo_file)

    zenodo[record_name] = record_id

    with open(HERE.joinpath("zenodo.json"), "w") as zenodo_file:
        json.dump(zenodo, zenodo_file, indent=2)


def get_zenodo_access_token(hostname=ZENODO_HOSTNAME):
    access_token = keyring.get_password("Zenodo", hostname)

    if access_token is None or access_token == "":
        print(
            "To upload files to Zenodo you need to provide a Zenodo "
            "access token. Please go to "
            f"https://{hostname}/account/settings/applications/tokens/new "
            "login to Zenodo and create a new access token.\n"
            "When creating the access token use the scopes `deposit:actions`, `deposit:write`, "
            "and `user:email`. Once you have your token copy it into the prompt below "
            "to continue with the upload."
        )
        access_token = getpass.getpass()
        set_zenodo_access_token(access_token, hostname)

    return access_token


def set_zenodo_access_token_cli(args):
    set_zenodo_access_token(args.token)


def set_zenodo_access_token(access_token, hostname=ZENODO_HOSTNAME):
    keyring.set_password("Zenodo", hostname, access_token)



# Copyright (C) 2018 Cancer Care Associates

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List

from pymedphys._imports import numpy as np
from pymedphys._imports import pandas as pd

from .constants import CONFIG
from .header import determine_header_length

GROUPING_OPTIONS = {
    "integrity_v3": {"line_grouping": 700, "linac_state_codes_column": 2},
    "integrity_v4": {"line_grouping": 708, "linac_state_codes_column": 6},
    "unity_experimental": {"line_grouping": 700, "linac_state_codes_column": 6},
}

# LINE_GROUPING_OPTIONS = {
#     item["line_grouping"]: item["linac_state_codes_column"]
#     for _, item in GROUPING_OPTIONS.items()
# }


def decode_rows(
    trf_table_contents,
    input_line_grouping=None,
    input_linac_state_codes_column=None,
    reference_state_code_keys=None,
):
    table_byte_length = len(trf_table_contents)

    if input_line_grouping is not None or input_linac_state_codes_column is not None:
        if input_line_grouping is None or input_linac_state_codes_column is None:
            raise ValueError(
                "If customising line grouping, need to provide both "
                "`input_line_grouping` and `input_linac_state_codes_column`"
            )

        grouping_options = {
            "custom": {
                "line_grouping": input_line_grouping,
                "linac_state_codes_column": input_linac_state_codes_column,
            }
        }
    else:
        grouping_options = GROUPING_OPTIONS

    possible_groupings = {}

    for key, item in grouping_options.items():
        grouping_option = item["line_grouping"]
        linac_state_codes_column = item["linac_state_codes_column"]

        if table_byte_length / grouping_option == table_byte_length // grouping_option:
            possible_groupings[key] = (grouping_option, linac_state_codes_column)

    if not possible_groupings:
        raise ValueError("Unexpected number of bytes within file.")

    if reference_state_code_keys is None:
        reference_state_codes = set(
            np.array(list(CONFIG["linac_state_codes"].keys())).astype(int)
        )
    else:
        reference_state_codes = set(reference_state_code_keys)

    decoded_results = []
    possible_column_adjustment_key = []
    for key, (line_grouping, linac_state_codes_column) in possible_groupings.items():
        rows = [
            trf_table_contents[i : i + line_grouping]
            for i in range(0, len(trf_table_contents), line_grouping)
        ]

        result = decode_table_data(rows, line_grouping)
        tentative_state_codes = result[  # pylint: disable = unsubscriptable-object
            :, linac_state_codes_column
        ]

        if set(tentative_state_codes).issubset(reference_state_codes):
            decoded_results.append(decode_table_data(rows, line_grouping))
            possible_column_adjustment_key.append(key)

    if not decoded_results:
        raise ValueError("Decoded table didn't pass shape test")

    if len(decoded_results) > 1:
        raise ValueError("Can't determine version of trf file from table shape")

    decoded_rows = decoded_results[0]
    column_adjustment_key = possible_column_adjustment_key[0]

    return decoded_rows, column_adjustment_key


def decode_rows_from_file(filepath):
    with open(filepath, "rb") as file:
        trf_contents = file.read()

    header_length = determine_header_length(trf_contents)
    trf_table_contents = trf_contents[header_length::]

    decoded_rows, _ = decode_rows(trf_table_contents)

    return decoded_rows


def get_base_column_names():
    return CONFIG["column_names"]


def get_column_names(column_adjustment_key):
    column_names = get_base_column_names()

    if column_adjustment_key == "integrity_v3":
        return column_names

    filler_columns = [f"unknown{item}" for item in range(1, 5)]

    column_names = filler_columns + column_names

    if column_adjustment_key == "integrity_v4":
        return column_names

    if column_adjustment_key != "unity_experimental":
        raise ValueError("Unexpected `column_adjustment_key`")

    column_names = [item for item in column_names if "Dlg" not in item]

    return column_names


def decode_trf_table(trf_table_contents):
    decoded_rows, column_adjustment_key = decode_rows(trf_table_contents)

    column_names = get_column_names(column_adjustment_key)

    number_of_columns = len(decoded_rows[0, :])

    if len(column_names) != number_of_columns:
        raise ValueError("Columns names don't agree with number of columns")

    table_dataframe = create_dataframe(
        decoded_rows, column_names, CONFIG["time_increment"]
    )

    convert_data_table(
        table_dataframe, CONFIG["linac_state_codes"], CONFIG["wedge_codes"]
    )

    return table_dataframe


def decode_data_item(row, group, byteorder) -> int:
    """Converts the bytes of data items into an integer."""
    return int.from_bytes(row[group], byteorder=byteorder)


def decode_column(raw_table_rows: List[str], column_number: int):
    """Decode all of the items in a given column."""
    grouping = 2
    i = column_number * grouping
    group = slice(i, i + grouping)
    byteorder = "little"

    column = np.array(
        [decode_data_item(row, group, byteorder) for row in raw_table_rows]
    )

    return column


def decode_table_data(raw_table_rows: List[str], line_grouping):
    """Decode the table into integer values."""

    result = []
    for column_number in range(0, line_grouping // 2):
        result.append(decode_column(raw_table_rows, column_number))

    return np.array(result).T


def create_dataframe(data, column_names, time_increment):
    """Converts the provided data into a pandas dataframe."""
    dataframe = pd.DataFrame(data=data, columns=column_names)
    dataframe.index = np.round(dataframe.index * time_increment, 2)

    return dataframe


def convert_numbers_to_string(name, lookup, column):
    dtype = np.array([item for _, item in lookup.items()]).dtype
    result = np.empty_like(column).astype(dtype)  # pylint: disable = no-member
    result[:] = ""

    for i, item in lookup.items():
        result[column.values == int(i)] = item

    if np.any(result == ""):
        print(lookup)
        print(np.where(result == ""))
        print(column[result == ""].values)
        unconverted_entries = np.unique(column[result == ""])
        raise ValueError(
            "The conversion lookup list for converting {} is incomplete. "
            "The following data numbers were not converted:\n"
            "{}\n"
            "Please update the trf2csv conversion script to include these "
            "in its definitions.".format(name, unconverted_entries)
        )

    return result


def convert_linac_state_codes(dataframe, linac_state_codes):
    name = "linac state"
    key = "Linac State/Actual Value (None)"
    dataframe[key] = convert_numbers_to_string(name, linac_state_codes, dataframe[key])


def convert_wedge_codes(dataframe, wedge_codes):
    name = "wedge"
    key = "Wedge Position/Actual Value (None)"
    dataframe[key] = convert_numbers_to_string(name, wedge_codes, dataframe[key])


def apply_negative(column):
    result = np.ones_like(column).astype(np.float64) * np.nan
    negative_values = column.values > 2 ** 15

    result[negative_values] = column[negative_values] - 2 ** 16
    result[np.invert(negative_values)] = column[np.invert(negative_values)]

    if np.any(np.isnan(result)):
        raise ValueError("Not all column values were converted")

    return result


def convert_applying_negative(dataframe):
    keys = [
        "Control point/Actual Value (None)",
        "Table Isocentric/Scaled Actual (deg)",
        "Table Isocentric/Positional Error (deg)",
    ]

    for key in keys:
        dataframe[key] = apply_negative(dataframe[key])


def negative_and_divide_by_10(column):
    result = apply_negative(column)
    result = result / 10

    return result


def convert_negative_and_divide_by_10(dataframe):
    keys = [
        "Step Dose/Actual Value (Mu)",
        "Step Gantry/Scaled Actual (deg)",
        "Step Gantry/Positional Error (deg)",
        "Step Collimator/Scaled Actual (deg)",
        "Step Collimator/Positional Error (deg)",
    ]

    for key in keys:
        dataframe[key] = negative_and_divide_by_10(dataframe[key])


def convert_remaining(dataframe):
    # The base column names are used here as they are presumed to be unchanging.
    # Should ever the order or contents of the base configured 'column_names'
    # change, this logic here will need to be changed.
    base_column_names = get_base_column_names()

    for key in base_column_names[14:30]:
        try:
            dataframe[key] = negative_and_divide_by_10(dataframe[key])
        except KeyError:
            if "Dlg" in key:
                # Unity logfile do not have a "Dlg" record
                pass
            else:
                raise

    # Previously a bug crept in due to this choice of logic. When the
    # decoding was adjusted to support Integrity 4 four extra
    # columns were added. This resulted in this logic being applied to
    # the wrong columns (offset by four).
    for key in base_column_names[30:110]:
        if "Leaf" not in key or "Y2" not in key or "Scaled Actual" not in key:
            raise ValueError("Y2 Leaf Keys were not in their expected positions.")

        # Y2 leaves need to be multiplied by -1
        dataframe[key] = -negative_and_divide_by_10(dataframe[key])

    for key in base_column_names[110::]:
        if "Leaf" not in key:
            raise ValueError(
                "The remaining leaf columns were not in their "
                f"expected positions. Key found was `{key}`."
            )
        dataframe[key] = negative_and_divide_by_10(dataframe[key])


def convert_data_table(dataframe, linac_state_codes, wedge_codes):
    convert_linac_state_codes(dataframe, linac_state_codes)
    convert_wedge_codes(dataframe, wedge_codes)

    convert_applying_negative(dataframe)
    convert_negative_and_divide_by_10(dataframe)
    convert_remaining(dataframe)



# Copyright (C) 2019 Cancer Care Associates

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from glob import glob


def wildcard_file_resolution(glob_search_string):
    filepaths = glob(glob_search_string)
    if len(filepaths) < 1:
        raise FileNotFoundError("No file found that matches the provided path")

    if len(filepaths) > 1:
        raise TypeError("More than one file found that matches the search string")

    found_filepath = filepaths[0]

    return found_filepath



# Copyright (C) 2019 Simon Biggs

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from .dicom import dicom_cli
from .pinnacle import pinnacle_cli
from .quickcheck import quickcheck_cli


def experimental_cli(subparsers):
    experimental_parser = subparsers.add_parser(
        "experimental", help="Experimental tools."
    )
    experimental_subparsers = experimental_parser.add_subparsers(dest="experimental")

    dicom_cli(experimental_subparsers)
    pinnacle_cli(experimental_subparsers)
    quickcheck_cli(experimental_subparsers)

    return experimental_parser



# Copyright (C) 2019 Simon Biggs

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pytest
from hypothesis import given, settings
from hypothesis.strategies import floats

import numpy as np

import pymedphys
import pymedphys._wlutz.findbb
import pymedphys._wlutz.imginterp
import pymedphys._wlutz.iview


@pytest.fixture
def test_field():
    image_path = pymedphys.data_path("wlutz_image.png", check_hash=False)
    (
        x,
        y,
        img,
    ) = pymedphys._wlutz.iview.iview_image_transform_from_path(  # pylint:disable = protected-access
        image_path
    )
    field = pymedphys._wlutz.imginterp.create_interpolated_field(  # pylint:disable = protected-access
        x, y, img
    )

    return field


@settings(deadline=None, max_examples=10)
@given(bb_centre_x_deviation=floats(-5, 5), bb_centre_y_deviation=floats(-5, 5))
def test_minimise_bb(
    bb_centre_x_deviation,
    bb_centre_y_deviation,
    test_field,  # pylint: disable = redefined-outer-name
):

    bb_diameter = 8

    reference_bb_centre = [1.47, -1.39]
    centre_to_test = [
        reference_bb_centre[0] + bb_centre_x_deviation,
        reference_bb_centre[1] + bb_centre_y_deviation,
    ]

    vectorised_to_minimise = pymedphys._wlutz.findbb.create_bb_to_minimise(  # pylint:disable = protected-access
        test_field, bb_diameter
    )
    simple_to_minimise = pymedphys._wlutz.findbb.create_bb_to_minimise_simple(  # pylint:disable = protected-access
        test_field, bb_diameter
    )

    assert np.allclose(
        vectorised_to_minimise(centre_to_test),
        simple_to_minimise(centre_to_test),
        atol=0.001,
    )



# Copyright (C) 2018 Matthew Jennings

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A test suite for the DICOM RT Dose toolbox."""

import json
from os.path import abspath, dirname
from os.path import join as pjoin
from zipfile import ZipFile

import pytest

import numpy as np

import pydicom

import pymedphys
from pymedphys._data import download
from pymedphys._dicom.collection import DicomDose
from pymedphys._dicom.dose import require_patient_orientation

from . import test_coords

HERE = dirname(abspath(__file__))
DATA_DIRECTORY = pjoin(HERE, "data", "dose")
ORIENTATIONS_SUPPORTED = ["FFDL", "FFDR", "FFP", "FFS", "HFDL", "HFDR", "HFP", "HFS"]


@pytest.mark.pydicom
def test_dicom_dose_constancy():
    wedge_basline_filename = "wedge_dose_baseline.json"

    baseline_dicom_dose_dict_zippath = download.get_file_within_data_zip(
        "dicom_dose_test_data.zip", "lfs-wedge_dose_baseline.zip"
    )
    test_dicom_dose_filepath = download.get_file_within_data_zip(
        "dicom_dose_test_data.zip", "RD.wedge.dcm"
    )

    test_dicom_dose = DicomDose.from_file(test_dicom_dose_filepath)

    with ZipFile(baseline_dicom_dose_dict_zippath, "r") as zip_ref:
        with zip_ref.open(wedge_basline_filename) as a_file:
            expected_dicom_dose_dict = json.load(a_file)

    assert np.allclose(
        test_dicom_dose.values, np.array(expected_dicom_dose_dict["values"])
    )
    assert test_dicom_dose.units == expected_dicom_dose_dict["units"]
    assert np.allclose(test_dicom_dose.x, np.array(expected_dicom_dose_dict["x"]))
    assert np.allclose(test_dicom_dose.y, np.array(expected_dicom_dose_dict["y"]))
    assert np.allclose(test_dicom_dose.z, np.array(expected_dicom_dose_dict["z"]))
    assert np.allclose(
        test_dicom_dose.coords, np.array(expected_dicom_dose_dict["coords"])
    )


@pytest.mark.pydicom
def test_require_patient_orientation():
    test_ds_dict = {
        key: pydicom.dcmread(test_coords.get_data_file(key))
        for key in ORIENTATIONS_SUPPORTED
    }

    ds_no_orient = pydicom.dcmread(
        str(pymedphys.data_path("example_structures.dcm")), force=True
    )

    test_ds_dict["no orient"] = ds_no_orient

    test_orientations = ("HFS", "HFP", "FFS", "FFP")

    for orient, ds in test_ds_dict.items():

        for test_orient in test_orientations:
            if orient == test_orient:
                require_patient_orientation(ds, test_orient)

            elif orient == "no orient":
                with pytest.raises(AttributeError):
                    require_patient_orientation(ds, test_orient)

            else:
                with pytest.raises(ValueError):
                    require_patient_orientation(ds, test_orient)



import functools
import re

from . import mappings

DATE_PATTERN = re.compile(rb"\d\d\d\d-\d\d-\d\d\d\d:\d\d:\d\d")


def get_data_points(data):
    date_index = [m.span() for m in DATE_PATTERN.finditer(data)]
    start_points = [span[0] - 8 for span in date_index]

    end_points = start_points[1::] + [None]

    data_points = [data[start:end] for start, end in zip(start_points, end_points)]
    return data_points


@functools.lru_cache()
def get_coll_regex(label, number):
    header = rb"0\xb8\x00DS\x00R.\x00\x00\x00" + label + b"\n"
    item = rb"0\x1c\x01DS\x00R.\x00\x00\x00(-?\d+\.\d+)"

    regex = re.compile(header + b"\n".join([item] * number))
    return regex


def extract_coll(data, label, number):
    regex = get_coll_regex(label, number)

    match = regex.search(data)
    span = match.span()

    data = data[0 : span[0]] + data[span[1] + 1 : :]
    items = [float(item) for item in match.groups()]

    return data, items


@functools.lru_cache()
def get_extraction_regex(key):
    regex = re.compile(
        rb"""[0\x00pP]""" + key + rb""".\x00\x00\x00([,\-'"a-zA-Z0-9 \.-]+)"""
    )
    return regex


def extract(data, label):
    key, this_type, where = mappings.ICOM[label]

    if where == "all":
        return extract_all(data, key, this_type)

    if where == "first":
        return extract_first(data, key, this_type)

    raise ValueError("Unexpected value for where")


def extract_all(data, key, this_type):
    result = []
    while True:
        data, a_result = extract_first(data, key, this_type)
        if a_result is not None:
            result.append(a_result)
        else:
            break

    return data, result


def extract_first(data, key, this_type):
    regex = get_extraction_regex(key)
    match = regex.search(data)

    try:
        span = match.span()
    except AttributeError:
        return data, None

    data = data[0 : span[0]] + data[span[1] + 1 : :]

    result = match.group(1)
    if result == b"-32767":
        data, result = extract_first(data, key, this_type)
    elif this_type is str:
        result = result.decode()
    else:
        result = this_type(result)

    return data, result



# Copyright (C) 2019 Simon Biggs

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
from copy import deepcopy

import pytest

import pydicom.filebase

from pymedphys._dicom.collection import DicomBase
from pymedphys._dicom.create import dicom_dataset_from_dict


@pytest.mark.pydicom
def test_copy():
    dont_change_string = "don't change me"
    to_be_changed_string = "do change me"

    new_manufactuer = "george"

    dataset_to_be_copied = dicom_dataset_from_dict({"Manufacturer": dont_change_string})

    dataset_to_be_viewed = dicom_dataset_from_dict(
        {"Manufacturer": to_be_changed_string}
    )

    dicom_base_copy = DicomBase(dataset_to_be_copied)
    dicom_base_view = DicomBase(dataset_to_be_viewed, copy=False)

    dicom_base_copy.dataset.Manufacturer = new_manufactuer
    dicom_base_view.dataset.Manufacturer = new_manufactuer

    assert dataset_to_be_copied.Manufacturer == dont_change_string
    assert dataset_to_be_viewed.Manufacturer == new_manufactuer


@pytest.mark.pydicom
def test_anonymise():
    expected_dataset = dicom_dataset_from_dict(
        {"Manufacturer": "PyMedPhys", "PatientName": "Anonymous"}
    )

    dicom = DicomBase.from_dict(
        {"Manufacturer": "PyMedPhys", "PatientName": "Python^Monte"}
    )

    dicom.anonymise(inplace=True)

    assert dicom.dataset == expected_dataset


@pytest.mark.pydicom
def test_to_and_from_file():
    temp_file = io.BytesIO()

    dicom = DicomBase.from_dict(
        {"Manufacturer": "PyMedPhys", "PatientName": "Python^Monte"}
    )

    dicom.to_file(temp_file)

    new_dicom = DicomBase.from_file(temp_file)

    # TODO: Without the str this was passing locally but not on CI. Further
    # investigation needed.
    assert new_dicom == dicom


@pytest.mark.pydicom
def test_equal():
    dicom1 = DicomBase.from_dict(
        {"Manufacturer": "PyMedPhys", "PatientName": "Python^Monte"}
    )
    dicom2 = DicomBase.from_dict(
        {"Manufacturer": "PyMedPhys", "PatientName": "Python^Monte"}
    )
    assert dicom1 == dicom2  # Equality from dict

    try:
        fp1 = pydicom.filebase.DicomBytesIO()
        dicom1.to_file(fp1)
        fp2 = pydicom.filebase.DicomBytesIO()
        dicom2.to_file(fp2)

        dicom1_from_file = DicomBase.from_file(fp1)
        dicom2_from_file = DicomBase.from_file(fp2)
        # Equality from file (implicitly also from dataset)
        assert dicom1_from_file == dicom2_from_file

        dicom1_from_file.dataset.PatientName = "test^PatientName change"
        assert dicom1_from_file != dicom2_from_file  # Negative case

        dicom1_from_file.dataset.PatientName = "Python^Monte"
        assert dicom1_from_file == dicom2_from_file  # Equality post re-assignment

        dicom1_from_file_copied = deepcopy(dicom1_from_file)
        assert dicom1_from_file == dicom1_from_file_copied  # Equality from deepcopy
    finally:
        fp1.close()
        fp2.close()



# Copyright (C) 2018 Cancer Care Associates

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Decodes trf file.
"""

import os  # pylint: disable = unused-import
from typing import Any, BinaryIO, Tuple, Union, cast  # pylint: disable = unused-import

from pymedphys._imports import pandas as pd

from .header import Header, decode_header
from .partition import split_into_header_table
from .table import decode_trf_table

path_or_binary_file = Union[BinaryIO, "os.PathLike[Any]"]


def trf2pandas(trf: path_or_binary_file) -> Tuple["pd.DataFrame", "pd.DataFrame"]:
    binary_file_trf = cast(BinaryIO, trf)
    path_like_trf = cast("os.PathLike[Any]", trf)

    try:
        binary_file_trf.seek(0)
        trf_contents = binary_file_trf.read()
    except AttributeError:
        with open(path_like_trf, "rb") as f:
            trf_contents = f.read()

    trf_header_contents, trf_table_contents = split_into_header_table(trf_contents)

    header_dataframe = header_as_dataframe(trf_header_contents)

    table_dataframe = decode_trf_table(trf_table_contents)

    return header_dataframe, table_dataframe


read_trf = trf2pandas


def header_as_dataframe(trf_header_contents):
    header = decode_header(trf_header_contents)

    return pd.DataFrame([header], columns=Header._fields)


def decode_trf(filepath):
    """DEPRECATED
    """
    _, table_dataframe = trf2pandas(filepath)

    return table_dataframe



# Copyright (C) 2020 Stuart Swerdloff, Simon Biggs
# Copyright (C) 2018 Matthew Jennings, Simon Biggs
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import logging

from pymedphys._imports import pydicom

from pymedphys._dicom.constants import PYMEDPHYS_ROOT_UID

REPLACEMENT_TO_VR_MAP = {
    "Anonymous": ["AE", "LO", "LT", "PN", "SH", "ST"],
    "100Y": ["AS"],
    "ANON": ["CS"],
    "20190303": ["DA"],
    "12345678.9": ["DS"],
    "20190303000900.000000": ["DT"],
    (0).to_bytes(2, "little"): ["OB", "OB or OW", "OW"],
    "000900.000000": ["TM"],
    PYMEDPHYS_ROOT_UID: ["UI"],
    12345: ["US"],
}

VR_TO_REPLACEMENT_MAP = {}
for value, keys in REPLACEMENT_TO_VR_MAP.items():
    for key in keys:
        VR_TO_REPLACEMENT_MAP[key] = value


def _get_vr_anonymous_hardcode_replacement_value(current_value, value_representation):
    """A single dispatch function that is used for any VR with the current_value of the element ignored
    This is the default for the replacement strategy
    """
    if current_value is None:
        # TODO: Decide if we actually want to be doing this. I haven't
        # been the biggest fan of this outcome.
        logging.debug("Replacing empty value with hardcoded anonymisation value")

    if value_representation == "SQ":
        # This is defined in here instead of within REPLACEMENT_TO_VR_MAP so that
        # pydicom does not get called during library import. This is to support
        # running other parts of pymedphys in environments where pydicom is not
        # installed.
        return [pydicom.Dataset()]

    return VR_TO_REPLACEMENT_MAP[value_representation]


_keys = list(VR_TO_REPLACEMENT_MAP.keys())
_keys.append("SQ")

ANONYMISATION_HARDCODE_DISPATCH = {
    key: functools.partial(
        _get_vr_anonymous_hardcode_replacement_value, value_representation=key
    )
    for key in _keys
}



# Copyright (C) 2019 Simon Biggs

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=C0103,C1801

from pymedphys._imports import numpy as np
from pymedphys._imports import plt

from pymedphys._utilities.constants import AGILITY
from pymedphys._utilities.controlpoints import remove_irrelevant_control_points

from .plt import pcolormesh_grid

__DEFAULT_LEAF_PAIR_WIDTHS = AGILITY
__DEFAULT_GRID_RESOLUTION = 1
__DEFAULT_MAX_LEAF_GAP = 400
__DEFAULT_MIN_STEP_PER_PIXEL = 10


def calc_metersetmap(
    mu,
    mlc,
    jaw,
    grid_resolution=None,
    max_leaf_gap=None,
    leaf_pair_widths=None,
    min_step_per_pixel=None,
):
    """Determine the MetersetMap.

    Both jaw and mlc positions are defined in bipolar format for each control
    point. A negative value indicates travel over the isocentre. All positional
    arguments are defined at the isocentre projection with the units of mm.

    Parameters
    ----------
    mu : numpy.ndarray
        1-D array containing an MU value for each control point.
    mlc : numpy.ndarray
        3-D array containing the MLC positions

            | axis 0: control point
            | axis 1: mlc pair
            | axis 2: leaf bank

    jaw : numpy.ndarray
        2-D array containing the jaw positions.

            | axis 0: control point
            | axis 1: diaphragm

    grid_resolution : float, optional
        The calc grid resolution. Defaults to 1 mm.

    max_leaf_gap : float, optional
        The maximum possible distance between opposing leaves. Defaults to
        400 mm.

    leaf_pair_widths : tuple, optional
        The widths of each leaf pair in the
        MLC limiting device. The number of entries in the tuples defines
        the number of leaf pairs. Each entry itself defines that particular
        leaf pair width. Defaults to 80 leaf pairs each 5 mm wide.

    min_step_per_pixel : int, optional
        The minimum number of time steps
        used per pixel for each control point. Defaults to 10.

    Returns
    -------
    metersetmap : numpy.ndarray
        2-D array containing the calculated metersetmap.

            | axis 0: jaw direction
            | axis 1: mlc direction

    Examples
    --------
    >>> import numpy as np
    >>> import pymedphys
    >>>
    >>> leaf_pair_widths = (5, 5, 5)
    >>> max_leaf_gap = 10
    >>> mu = np.array([0, 2, 5, 10])
    >>> mlc = np.array([
    ...     [
    ...         [1, 1],
    ...         [2, 2],
    ...         [3, 3]
    ...     ],
    ...     [
    ...         [2, 2],
    ...         [3, 3],
    ...         [4, 4]
    ...     ],
    ...     [
    ...         [-2, 3],
    ...         [-2, 4],
    ...         [-2, 5]
    ...     ],
    ...     [
    ...         [0, 0],
    ...         [0, 0],
    ...         [0, 0]
    ...     ]
    ... ])
    >>> jaw = np.array([
    ...     [7.5, 7.5],
    ...     [7.5, 7.5],
    ...     [-2, 7.5],
    ...     [0, 0]
    ... ])
    >>>
    >>> grid = pymedphys.metersetmap.grid(
    ...    max_leaf_gap=max_leaf_gap, leaf_pair_widths=leaf_pair_widths)
    >>> grid['mlc']
    array([-5., -4., -3., -2., -1.,  0.,  1.,  2.,  3.,  4.,  5.])
    >>>
    >>> grid['jaw']
    array([-8., -7., -6., -5., -4., -3., -2., -1.,  0.,  1.,  2.,  3.,  4.,
            5.,  6.,  7.,  8.])
    >>>
    >>> metersetmap = pymedphys.metersetmap.calculate(
    ...    mu, mlc, jaw, max_leaf_gap=max_leaf_gap,
    ...    leaf_pair_widths=leaf_pair_widths)
    >>> pymedphys.metersetmap.display(grid, metersetmap)
    >>>
    >>> np.round(metersetmap, 1)
    array([[0. , 0. , 0. , 0. , 0. , 0. , 0. , 0. , 0. , 0. , 0. ],
           [0. , 0. , 0. , 0.3, 1.9, 2.2, 1.9, 0.4, 0. , 0. , 0. ],
           [0. , 0. , 0. , 0.4, 2.2, 2.5, 2.2, 0.6, 0. , 0. , 0. ],
           [0. , 0. , 0. , 0.4, 2.4, 2.8, 2.5, 0.8, 0. , 0. , 0. ],
           [0. , 0. , 0. , 0.4, 2.5, 3.1, 2.8, 1. , 0. , 0. , 0. ],
           [0. , 0. , 0. , 0.4, 2.5, 3.4, 3.1, 1.3, 0. , 0. , 0. ],
           [0. , 0. , 0.4, 2.3, 3.2, 3.7, 3.7, 3.5, 1.6, 0. , 0. ],
           [0. , 0. , 0.4, 2.3, 3.2, 3.8, 4. , 3.8, 1.9, 0.1, 0. ],
           [0. , 0. , 0.4, 2.3, 3.2, 3.8, 4.3, 4.1, 2.3, 0.1, 0. ],
           [0. , 0. , 0.4, 2.3, 3.2, 3.9, 5.2, 4.7, 2.6, 0.2, 0. ],
           [0. , 0. , 0.4, 2.3, 3.2, 3.8, 5.4, 6.6, 3.8, 0.5, 0. ],
           [0. , 0.3, 2.2, 3. , 3.5, 4. , 5.1, 7.5, 6.7, 3.9, 0.5],
           [0. , 0.3, 2.2, 3. , 3.5, 4. , 4.7, 6.9, 6.7, 3.9, 0.5],
           [0. , 0.3, 2.2, 3. , 3.5, 4. , 4.5, 6.3, 6.4, 3.9, 0.5],
           [0. , 0.3, 2.2, 3. , 3.5, 4. , 4.5, 5.6, 5.7, 3.8, 0.5],
           [0. , 0.3, 2.2, 3. , 3.5, 4. , 4.5, 5.1, 5.1, 3.3, 0.5],
           [0. , 0. , 0. , 0. , 0. , 0. , 0. , 0. , 0. , 0. , 0. ]])


    MetersetMap from a Mosaiq record

    >>> import pymedphys
    >>>
    >>> def metersetmap_from_mosaiq(msq_server_name, field_id):
    ...     with pymedphys.mosaiq.connect(msq_server_name) as cursor:
    ...         delivery = pymedphys.Delivery.from_mosaiq(cursor, field_id)
    ...
    ...     grid = pymedphys.metersetmap.grid()
    ...     metersetmap = delivery.metersetmap()
    ...     pymedphys.metersetmap.display(grid, metersetmap)
    >>>
    >>> metersetmap_from_mosaiq('a_server_name', 11111) # doctest: +SKIP


    MetersetMap from a logfile at a given filepath

    >>> import pymedphys
    >>>
    >>> def metersetmap_from_logfile(filepath):
    ...     delivery_data = Delivery.from_logfile(filepath)
    ...     metersetmap = Delivery.metersetmap()
    ...
    ...     grid = pymedphys.metersetmap.grid()
    ...     pymedphys.metersetmap.display(grid, metersetmap)
    >>>
    >>> metersetmap_from_logfile(r"a/path/goes/here")  # doctest: +SKIP

    """

    if grid_resolution is None:
        grid_resolution = __DEFAULT_GRID_RESOLUTION

    if max_leaf_gap is None:
        max_leaf_gap = __DEFAULT_MAX_LEAF_GAP

    if leaf_pair_widths is None:
        leaf_pair_widths = __DEFAULT_LEAF_PAIR_WIDTHS

    if min_step_per_pixel is None:
        min_step_per_pixel = __DEFAULT_MIN_STEP_PER_PIXEL

    divisibility_of_max_leaf_gap = np.array(max_leaf_gap / 2 / grid_resolution)
    max_leaf_gap_is_divisible = (
        divisibility_of_max_leaf_gap.astype(int) == divisibility_of_max_leaf_gap
    )

    if not max_leaf_gap_is_divisible:
        raise ValueError(
            "The grid resolution needs to be able to divide the max leaf gap exactly by"
            " four"
        )

    leaf_pair_widths = np.array(leaf_pair_widths)

    if not np.max(np.abs(mlc)) <= max_leaf_gap / 2:  # pylint: disable = unneeded-not
        first_failing_control_point = np.where(np.abs(mlc) > max_leaf_gap / 2)[0][0]

        raise ValueError(
            "The mlc should not travel further out than half the maximum leaf gap.\n"
            "The first failing control point has the following positions:\n"
            f"{np.array(mlc)[first_failing_control_point, :, :]}"
        )

    mu, mlc, jaw = remove_irrelevant_control_points(mu, mlc, jaw)

    full_grid = get_grid(max_leaf_gap, grid_resolution, leaf_pair_widths)

    metersetmap = np.zeros((len(full_grid["jaw"]), len(full_grid["mlc"])))

    for i in range(len(mu) - 1):
        control_point_slice = slice(i, i + 2, 1)
        current_mlc = mlc[control_point_slice, :, :]
        current_jaw = jaw[control_point_slice, :]
        delivered_mu = np.diff(mu[control_point_slice])

        grid, metersetmap_of_slice = calc_single_control_point(
            current_mlc,
            current_jaw,
            delivered_mu,
            leaf_pair_widths=leaf_pair_widths,
            grid_resolution=grid_resolution,
            min_step_per_pixel=min_step_per_pixel,
        )
        full_grid_metersetmap_of_slice = _convert_to_full_grid(
            grid, full_grid, metersetmap_of_slice
        )

        metersetmap += full_grid_metersetmap_of_slice

    return metersetmap


def calc_single_control_point(
    mlc,
    jaw,
    delivered_mu=1,
    leaf_pair_widths=__DEFAULT_LEAF_PAIR_WIDTHS,
    grid_resolution=__DEFAULT_GRID_RESOLUTION,
    min_step_per_pixel=__DEFAULT_MIN_STEP_PER_PIXEL,
):
    """Calculate the MetersetMap for a single control point.

    Examples
    --------
    >>> from pymedphys._imports import numpy as np
    >>> from pymedphys._metersetmap.metersetmap import (
    ...     calc_single_control_point, display_metersetmap)
    >>>
    >>> leaf_pair_widths = (2, 2)
    >>> mlc = np.array([
    ...     [
    ...         [1, 1],
    ...         [2, 2],
    ...     ],
    ...     [
    ...         [2, 2],
    ...         [3, 3],
    ...     ]
    ... ])
    >>> jaw = np.array([
    ...     [1.5, 1.2],
    ...     [1.5, 1.2]
    ... ])
    >>> grid, metersetmap = calc_single_control_point(
    ...     mlc, jaw, leaf_pair_widths=leaf_pair_widths)
    >>> display_metersetmap(grid, metersetmap)
    >>>
    >>> grid['mlc']
    array([-3., -2., -1.,  0.,  1.,  2.,  3.])
    >>>
    >>> grid['jaw']
    array([-1.5, -0.5,  0.5,  1.5])
    >>>
    >>> np.round(metersetmap, 2)
    array([[0.  , 0.07, 0.43, 0.5 , 0.43, 0.07, 0.  ],
           [0.  , 0.14, 0.86, 1.  , 0.86, 0.14, 0.  ],
           [0.14, 0.86, 1.  , 1.  , 1.  , 0.86, 0.14],
           [0.03, 0.17, 0.2 , 0.2 , 0.2 , 0.17, 0.03]])
    """

    leaf_pair_widths = np.array(leaf_pair_widths)
    leaf_division = leaf_pair_widths / grid_resolution

    if not np.all(leaf_division.astype(int) == leaf_division):
        raise ValueError(
            "The grid resolution needs to exactly divide every leaf pair width."
        )

    if (
        not np.max(np.abs(jaw))  # pylint: disable = unneeded-not
        <= np.sum(leaf_pair_widths) / 2
    ):
        raise ValueError(
            "The jaw should not travel further out than the maximum leaf limits. "
            f"Max travel was {np.max(np.abs(jaw))}"
        )

    (grid, grid_leaf_map, mlc) = _determine_calc_grid_and_adjustments(
        mlc, jaw, leaf_pair_widths, grid_resolution
    )

    positions = {
        "mlc": {
            1: (-mlc[0, :, 0], -mlc[1, :, 0]),  # left
            -1: (mlc[0, :, 1], mlc[1, :, 1]),  # right
        },
        "jaw": {
            1: (-jaw[0::-1, 0], -jaw[1::, 0]),  # bot
            -1: (jaw[0::-1, 1], jaw[1::, 1]),  # top
        },
    }

    time_steps = _calc_time_steps(positions, grid_resolution, min_step_per_pixel)
    blocked_by_device = _calc_blocked_by_device(
        grid, positions, grid_resolution, time_steps
    )
    device_open = _calc_device_open(blocked_by_device)
    mlc_open, jaw_open = _remap_mlc_and_jaw(device_open, grid_leaf_map)
    open_fraction = _calc_open_fraction(mlc_open, jaw_open)

    metersetmap = open_fraction * delivered_mu

    return grid, metersetmap


def single_mlc_pair(
    left_mlc,
    right_mlc,
    grid_resolution=__DEFAULT_GRID_RESOLUTION,
    min_step_per_pixel=__DEFAULT_MIN_STEP_PER_PIXEL,
):
    """Calculate the MetersetMap of a single leaf pair.

    Examples
    --------
    >>> from pymedphys._imports import numpy as np
    >>> import matplotlib.pyplot as plt
    >>>
    >>> from pymedphys._metersetmap.metersetmap import single_mlc_pair
    >>>
    >>> mlc_left = (-2.3, 3.1)  # (start position, end position)
    >>> mlc_right = (0, 7.7)
    >>>
    >>> x, metersetmap = single_mlc_pair(mlc_left, mlc_right)
    >>> fig = plt.plot(x, metersetmap, '-o')
    >>>
    >>> x
    array([-2., -1.,  0.,  1.,  2.,  3.,  4.,  5.,  6.,  7.,  8.])
    >>>
    >>> np.round(metersetmap, 3)
    array([0.064, 0.244, 0.408, 0.475, 0.53 , 0.572, 0.481, 0.352, 0.224,
           0.096, 0.004])
    """
    leaf_pair_widths = [grid_resolution]
    jaw = np.array(
        [
            [grid_resolution / 2, grid_resolution / 2],
            [grid_resolution / 2, grid_resolution / 2],
        ]
    )
    mlc = np.array([[[-left_mlc[0], right_mlc[0]]], [[-left_mlc[1], right_mlc[1]]]])

    grid, metersetmap = calc_single_control_point(
        mlc,
        jaw,
        leaf_pair_widths=leaf_pair_widths,
        grid_resolution=grid_resolution,
        min_step_per_pixel=min_step_per_pixel,
    )

    return grid["mlc"], metersetmap[0, :]


def calc_metersetmap_return_grid(
    mu,
    mlc,
    jaw,
    grid_resolution=__DEFAULT_GRID_RESOLUTION,
    max_leaf_gap=__DEFAULT_MAX_LEAF_GAP,
    leaf_pair_widths=__DEFAULT_LEAF_PAIR_WIDTHS,
    min_step_per_pixel=__DEFAULT_MIN_STEP_PER_PIXEL,
):
    """DEPRECATED. This is a temporary helper function to provide the old
    api.
    """

    leaf_pair_widths = np.array(leaf_pair_widths)
    metersetmap = calc_metersetmap(
        mu,
        mlc,
        jaw,
        grid_resolution=grid_resolution,
        max_leaf_gap=max_leaf_gap,
        leaf_pair_widths=leaf_pair_widths,
        min_step_per_pixel=min_step_per_pixel,
    )

    full_grid = get_grid(max_leaf_gap, grid_resolution, leaf_pair_widths)

    grid_xx, grid_yy = np.meshgrid(full_grid["mlc"], full_grid["jaw"])

    return grid_xx, grid_yy, metersetmap


def get_grid(
    max_leaf_gap=__DEFAULT_MAX_LEAF_GAP,
    grid_resolution=__DEFAULT_GRID_RESOLUTION,
    leaf_pair_widths=__DEFAULT_LEAF_PAIR_WIDTHS,
):
    """Get the MetersetMap grid for plotting purposes.

    Examples
    --------
    See `pymedphys.metersetmap.calculate`_.
    """

    leaf_pair_widths = np.array(leaf_pair_widths)

    grid = dict()

    grid["mlc"] = np.arange(
        -max_leaf_gap / 2, max_leaf_gap / 2 + grid_resolution, grid_resolution
    ).astype("float")

    _, top_of_reference_leaf = _determine_leaf_centres(leaf_pair_widths)
    grid_reference_position = _determine_reference_grid_position(
        top_of_reference_leaf, grid_resolution
    )

    # It might be better to use round instead of ceil here.
    total_leaf_widths = np.sum(leaf_pair_widths)
    top_grid_pos = (
        np.ceil((total_leaf_widths / 2 - grid_reference_position) / grid_resolution)
        * grid_resolution
        + grid_reference_position
    )

    bot_grid_pos = (
        grid_reference_position
        - np.ceil((total_leaf_widths / 2 + grid_reference_position) / grid_resolution)
        * grid_resolution
    )

    grid["jaw"] = np.arange(
        bot_grid_pos, top_grid_pos + grid_resolution, grid_resolution
    )

    return grid


def display_metersetmap_diff(
    grid, metersetmap_eval, metersetmap_ref, grid_resolution=None, colour_range=None
):
    cmap = "bwr"
    diff = metersetmap_eval - metersetmap_ref
    if colour_range is None:
        colour_range = np.max(np.abs(diff))

    # pylint: disable=invalid-unary-operand-type
    display_metersetmap(
        grid,
        diff,
        grid_resolution=grid_resolution,
        cmap=cmap,
        vmin=-colour_range,
        vmax=colour_range,
    )


def display_metersetmap(
    grid, metersetmap, grid_resolution=None, cmap=None, vmin=None, vmax=None
):
    """Prints a colour plot of the MetersetMap.

    Examples
    --------
    See `pymedphys.metersetmap.calculate`_.
    """
    if grid_resolution is None:
        grid_resolution = grid["mlc"][1] - grid["mlc"][0]

    x, y = pcolormesh_grid(grid["mlc"], grid["jaw"], grid_resolution)
    plt.pcolormesh(x, y, metersetmap, cmap=cmap, vmin=vmin, vmax=vmax)
    plt.colorbar()
    plt.title("MetersetMap")
    plt.xlabel("MLC direction (mm)")
    plt.ylabel("Jaw direction (mm)")
    plt.axis("equal")
    plt.gca().invert_yaxis()


def _calc_blocked_t(travel_diff, grid_resolution):
    blocked_t = np.ones_like(travel_diff) * np.nan

    fully_blocked = travel_diff <= -grid_resolution / 2
    fully_open = travel_diff >= grid_resolution / 2
    blocked_t[fully_blocked] = 1
    blocked_t[fully_open] = 0

    transient = ~fully_blocked & ~fully_open

    blocked_t[transient] = (
        -travel_diff[transient] + grid_resolution / 2
    ) / grid_resolution

    assert np.all(~np.isnan(blocked_t))

    return blocked_t


def _calc_time_steps(positions, grid_resolution, min_step_per_pixel):
    maximum_travel = []
    for _, value in positions.items():
        for _, (start, end) in value.items():
            maximum_travel.append(np.max(np.abs(end - start)))

    maximum_travel = np.max(maximum_travel)
    number_of_pixels = np.ceil(maximum_travel / grid_resolution)
    time_steps = number_of_pixels * min_step_per_pixel

    if time_steps < 10:
        time_steps = 10

    return time_steps


def _calc_blocked_by_device(grid, positions, grid_resolution, time_steps):
    blocked_by_device = {}

    for device, value in positions.items():
        blocked_by_device[device] = dict()

        for multiplier, (start, end) in value.items():
            dt = (end - start) / (time_steps - 1)
            travel = start[None, :] + np.arange(0, time_steps)[:, None] * dt[None, :]
            travel_diff = multiplier * (
                grid[device][None, None, :] - travel[:, :, None]
            )

            blocked_by_device[device][multiplier] = _calc_blocked_t(
                travel_diff, grid_resolution
            )

    return blocked_by_device


def _calc_device_open(blocked_by_device):
    device_open = {}

    for device, value in blocked_by_device.items():
        device_sum = np.sum(
            np.concatenate(
                [np.expand_dims(blocked, axis=0) for _, blocked in value.items()],
                axis=0,
            ),
            axis=0,
        )

        device_open[device] = 1 - device_sum

    return device_open


def _remap_mlc_and_jaw(device_open, grid_leaf_map):
    mlc_open = device_open["mlc"][:, grid_leaf_map, :]
    jaw_open = device_open["jaw"][:, 0, :]

    return mlc_open, jaw_open


def _calc_open_fraction(mlc_open, jaw_open):
    open_t = mlc_open * jaw_open[:, :, None]
    open_fraction = np.mean(open_t, axis=0)

    return open_fraction


def _determine_leaf_centres(leaf_pair_widths):
    total_leaf_widths = np.sum(leaf_pair_widths)
    leaf_centres = (
        np.cumsum(leaf_pair_widths) - leaf_pair_widths / 2 - total_leaf_widths / 2
    )

    reference_leaf_index = len(leaf_centres) // 2

    top_of_reference_leaf = (
        leaf_centres[reference_leaf_index] + leaf_pair_widths[reference_leaf_index] / 2
    )

    return leaf_centres, top_of_reference_leaf


def _determine_reference_grid_position(top_of_reference_leaf, grid_resolution):
    grid_reference_position = top_of_reference_leaf - grid_resolution / 2

    return grid_reference_position


def _determine_calc_grid_and_adjustments(mlc, jaw, leaf_pair_widths, grid_resolution):
    min_y = np.min(-jaw[:, 0])
    max_y = np.max(jaw[:, 1])

    leaf_centres, top_of_reference_leaf = _determine_leaf_centres(leaf_pair_widths)
    grid_reference_position = _determine_reference_grid_position(
        top_of_reference_leaf, grid_resolution
    )

    top_grid_pos = (
        np.round((max_y - grid_reference_position) / grid_resolution)
    ) * grid_resolution + grid_reference_position

    bot_grid_pos = (
        grid_reference_position
        - (np.round((-min_y + grid_reference_position) / grid_resolution))
        * grid_resolution
    )

    grid = dict()
    grid["jaw"] = np.arange(
        bot_grid_pos, top_grid_pos + grid_resolution, grid_resolution
    ).astype("float")

    grid_leaf_map = np.argmin(
        np.abs(grid["jaw"][:, None] - leaf_centres[None, :]), axis=1
    )

    adjusted_grid_leaf_map = grid_leaf_map - np.min(grid_leaf_map)

    leaves_to_be_calced = np.unique(grid_leaf_map)
    adjusted_mlc = mlc[:, leaves_to_be_calced, :]

    min_x = np.round(np.min(-adjusted_mlc[:, :, 0]) / grid_resolution) * grid_resolution
    max_x = np.round(np.max(adjusted_mlc[:, :, 1]) / grid_resolution) * grid_resolution

    grid["mlc"] = np.arange(min_x, max_x + grid_resolution, grid_resolution).astype(
        "float"
    )

    return grid, adjusted_grid_leaf_map, adjusted_mlc


def _convert_to_full_grid(grid, full_grid, metersetmap):
    grid_xx, grid_yy = np.meshgrid(grid["mlc"], grid["jaw"])
    full_grid_xx, full_grid_yy = np.meshgrid(full_grid["mlc"], full_grid["jaw"])

    xx_from, xx_to = np.where(
        np.abs(full_grid_xx[None, 0, :] - grid_xx[0, :, None]) < 0.0001
    )
    yy_from, yy_to = np.where(
        np.abs(full_grid_yy[None, :, 0] - grid_yy[:, 0, None]) < 0.0001
    )

    full_grid_metersetmap = np.zeros_like(full_grid_xx)
    full_grid_metersetmap[  # pylint: disable=unsupported-assignment-operation
        np.ix_(yy_to, xx_to)
    ] = metersetmap[np.ix_(yy_from, xx_from)]

    return full_grid_metersetmap



# Copyright (C) 2020 Simon Biggs

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import json
import pathlib

from pymedphys._imports import pydicom


def resolve_paths(root, paths_dict):
    root = pathlib.Path(root)
    resolved_paths_dict = {key: root.joinpath(item) for key, item in paths_dict.items()}

    return resolved_paths_dict


def read_uid_cache(data_path_root, uid_cache):
    ct_uid_to_structure_uid = uid_cache["ct_uid_to_structure_uid"]

    structure_uid_to_ct_uids = {}
    for ct_uid, structure_uid in ct_uid_to_structure_uid.items():
        try:
            structure_uid_to_ct_uids[structure_uid].append(ct_uid)
        except KeyError:
            structure_uid_to_ct_uids[structure_uid] = [ct_uid]

    return (
        resolve_paths(data_path_root, uid_cache["ct_image_paths"]),
        resolve_paths(data_path_root, uid_cache["structure_set_paths"]),
        ct_uid_to_structure_uid,
        structure_uid_to_ct_uids,
    )


def get_uid_cache(data_path_root, validate_cache=True):
    """Get inter-relationship maps between dicom files.

    Dicom files are to be placed within ``data_path_root/dicom``.
    A cache will be created at ``data_path_root/mappings/uid-cache.json``.

    Any change of the file structure within data_path_root/dicom will
    cause the cache to be flushed.
    """

    data_path_root = pathlib.Path(data_path_root)
    uid_cache_path = data_path_root.joinpath("mappings", "uid-cache.json")
    dcm_paths = list(data_path_root.rglob("dicom/**/*.dcm"))

    relative_paths = [str(path.relative_to(data_path_root)) for path in dcm_paths]

    try:
        with open(uid_cache_path) as f:
            uid_cache = json.load(f)
    except (FileNotFoundError, json.JSONDecodeError):
        uid_cache = {
            "ct_image_paths": {},
            "structure_set_paths": {},
            "ct_uid_to_structure_uid": {},
            "paths_when_run": [],
        }

    if validate_cache and set(uid_cache["paths_when_run"]) != set(relative_paths):
        dcm_headers = []
        for dcm_path in dcm_paths:
            dcm_headers.append(
                pydicom.read_file(
                    dcm_path,
                    force=True,
                    specific_tags=["SOPInstanceUID", "SOPClassUID", "StudyInstanceUID"],
                )
            )

        ct_image_paths = {
            str(header.SOPInstanceUID): str(path)
            for header, path in zip(dcm_headers, relative_paths)
            if header.SOPClassUID.name == "CT Image Storage"
        }

        structure_set_paths = {
            str(header.SOPInstanceUID): str(path)
            for header, path in zip(dcm_headers, relative_paths)
            if header.SOPClassUID.name == "RT Structure Set Storage"
        }

        ct_uid_to_study_instance_uid = {
            str(header.SOPInstanceUID): str(header.StudyInstanceUID)
            for header in dcm_headers
            if header.SOPClassUID.name == "CT Image Storage"
        }

        study_instance_uid_to_structure_uid = {
            str(header.StudyInstanceUID): str(header.SOPInstanceUID)
            for header in dcm_headers
            if header.SOPClassUID.name == "RT Structure Set Storage"
        }

        ct_uid_to_structure_uid = {
            ct_uid: study_instance_uid_to_structure_uid[study_uid]
            for ct_uid, study_uid in ct_uid_to_study_instance_uid.items()
        }

        uid_cache["ct_image_paths"] = ct_image_paths
        uid_cache["structure_set_paths"] = structure_set_paths
        uid_cache["ct_uid_to_structure_uid"] = ct_uid_to_structure_uid
        uid_cache["paths_when_run"] = relative_paths

        with open(uid_cache_path, "w") as f:
            json.dump(uid_cache, f)

    return read_uid_cache(data_path_root, uid_cache)


def get_structure_names_by_uids(structure_set_paths, names_map):
    structure_names_by_ct_uid = {}
    structure_names_by_structure_set_uid = {}

    for structure_set_uid, structure_set_path in structure_set_paths.items():
        structure_set = pydicom.read_file(
            structure_set_path,
            force=True,
            specific_tags=["ROIContourSequence", "StructureSetROISequence"],
        )

        number_to_name_map = {
            roi_sequence_item.ROINumber: names_map[roi_sequence_item.ROIName]
            for roi_sequence_item in structure_set.StructureSetROISequence
            if names_map[roi_sequence_item.ROIName] is not None
        }

        structure_names_by_structure_set_uid[structure_set_uid] = [
            item for _, item in number_to_name_map.items()
        ]

        for roi_contour_sequence_item in structure_set.ROIContourSequence:
            try:
                structure_name = number_to_name_map[
                    roi_contour_sequence_item.ReferencedROINumber
                ]
            except KeyError:
                continue

            for contour_sequence_item in roi_contour_sequence_item.ContourSequence:
                contour_imaging_sequence = contour_sequence_item.ContourImageSequence
                ct_uid = contour_imaging_sequence[0].ReferencedSOPInstanceUID

                try:
                    structure_names_by_ct_uid[ct_uid].add(structure_name)
                except KeyError:
                    structure_names_by_ct_uid[ct_uid] = set([structure_name])

    structure_names_by_ct_uid = {
        key: list(item) for key, item in structure_names_by_ct_uid.items()
    }

    return structure_names_by_ct_uid, structure_names_by_structure_set_uid


def get_cached_structure_names_by_uids(data_path_root, structure_set_paths, names_map):
    data_path_root = pathlib.Path(data_path_root)
    structure_names_mapping_cache_path = data_path_root.joinpath(
        "mappings", "structure-names-mapping-cache.json"
    )

    relative_structure_set_paths = {
        key: str(pathlib.Path(path).relative_to(data_path_root))
        for key, path in structure_set_paths.items()
    }

    try:
        with open(structure_names_mapping_cache_path) as f:
            structure_names_cache = json.load(f)
    except (FileNotFoundError, json.JSONDecodeError):
        structure_names_cache = {
            "structure_names_by_ct_uid": {},
            "structure_names_by_structure_set_uid": {},
            "structure_set_paths_when_run": {},
            "names_map_when_run": {},
        }

    cache_valid = (
        structure_names_cache["structure_set_paths_when_run"].keys()
        == relative_structure_set_paths.keys()
        and structure_names_cache["names_map_when_run"] == names_map
    )

    if not cache_valid:
        (
            structure_names_by_ct_uid,
            structure_names_by_structure_set_uid,
        ) = get_structure_names_by_uids(structure_set_paths, names_map)

        structure_names_cache["structure_names_by_ct_uid"] = structure_names_by_ct_uid
        structure_names_cache[
            "structure_names_by_structure_set_uid"
        ] = structure_names_by_structure_set_uid
        structure_names_cache[
            "structure_set_paths_when_run"
        ] = relative_structure_set_paths
        structure_names_cache["names_map_when_run"] = names_map

        with open(structure_names_mapping_cache_path, "w") as f:
            json.dump(structure_names_cache, f)

    structure_names_by_ct_uid = structure_names_cache["structure_names_by_ct_uid"]
    structure_names_by_structure_set_uid = structure_names_cache[
        "structure_names_by_structure_set_uid"
    ]

    return structure_names_by_ct_uid, structure_names_by_structure_set_uid



# -*- coding: utf-8 -*-
"""
Created on Fri Mar 13 14:29:35 2020

@author: D. Buckle
*******************************************************************************
Some operations to load and prepare plan data for calculation of plan metrics

openPlan - Opens a DICOM plan file and returns the whole data as 'ds'

getBeamType - extract the beam types from the plan as some metrics are calculated
    differently depending on whether they are VMAT.

getDeomgraphics - Extract the demographics information using the dicompyler-core
    dicomparser module

getmlcdata - Reads in the DICOM data from openPlan and extracts the MLC and
    jaws data per beam and per control point into individual dictionaries for
    easier visualisation and to make them easier to work with.
    Returns a tuple containing the MLC data and the jaw data stored in
    dictionaries. Each beam itself is a dictionary of numpy arrays for each CP

*******************************************************************************

"""

from pymedphys._imports import numpy as np
from pymedphys._imports import pydicom

from . import data as MLCdata

# from dicompylercore import dicomparser


def openPlan(plan):
    ds = pydicom.dcmread(plan)
    return ds


def getBeamType(ds, b):
    beamType = ds.BeamSequence[b].BeamType
    return beamType


# def getDemographics(file):
#    demographics = dicomparser.DicomParser.GetDemographics(file)
#    print(demographics)


def getMLCdata(ds):
    leafIndices = MLCdata.TrueBeam  # import leaf index values from MLC data file
    beamSequence = ds.BeamSequence
    beamNames = []  # getting the beam names
    mlcData = {}
    jawData = {}
    segWeightData = {}
    maxPositions = {}
    maxJawPos = {}
    maxSep = {}
    maxSepSum = {}
    for b in beamSequence:  # looping through the beams
        name = b.BeamName
        beamNames.append(name)
        cps = b.ControlPointSequence
        cp = len(cps)

        leafPositions = {}  # empty dictionary to place leaf positions into
        jawPositions = {}
        segWeight = {}
        maxA = np.zeros((60))
        maxB = np.zeros((60))
        maxJaws = np.zeros((4))
        maxPos = []
        for i in range(0, cp - 1):  # looping through control points
            if len(cps[i].BeamLimitingDevicePositionSequence) == 3:  # assymm jaws
                xJaws_ = cps[i].BeamLimitingDevicePositionSequence[0].LeafJawPositions
                yJaws_ = cps[i].BeamLimitingDevicePositionSequence[1].LeafJawPositions
                mlcPos = cps[i].BeamLimitingDevicePositionSequence[2].LeafJawPositions
            elif len(cps[i].BeamLimitingDevicePositionSequence) == 2:  # symmetric jaws
                xJaws_ = cps[i].BeamLimitingDevicePositionSequence[0].LeafJawPositions
                yJaws_ = cps[i].BeamLimitingDevicePositionSequence[0].LeafJawPositions
                mlcPos = cps[i].BeamLimitingDevicePositionSequence[1].LeafJawPositions
            leafPosA = list(map(float, mlcPos[0:60]))
            leafPosB = list(map(float, mlcPos[60:120]))
            leafPositions[i] = np.column_stack((leafIndices, leafPosA, leafPosB))
            xJaws = list(map(float, xJaws_))
            yJaws = list(map(float, yJaws_))
            positions = list((xJaws[0], xJaws[1], yJaws[0], yJaws[1]))
            indices = list(("x1", "x2", "y1", "y2"))
            jawPositions[i] = np.column_stack((indices, positions))

            if i == 0:
                segWeight[i] = float(cps[i + 1].CumulativeMetersetWeight)
            else:
                segWeight[i] = float(
                    cps[i + 1].CumulativeMetersetWeight
                    - cps[i].CumulativeMetersetWeight
                )

            for l in range(0, 60):
                if leafPosA[l] < maxA[l]:
                    maxA[l] = float(leafPosA[l])
                if leafPosB[l] > maxB[l]:
                    maxB[l] = float(leafPosB[l])

            for l in range(0, 4):
                if abs(positions[l]) > abs(maxJaws[l]):
                    maxJaws[l] = float(positions[l])

        string = str(f"{name}")
        maxPos = np.column_stack((maxA, maxB))
        mlcData[string] = leafPositions
        jawData[string] = jawPositions
        segWeightData[string] = segWeight
        maxPositions[string] = maxPos
        maxJawPos[string] = maxJaws

        _ = list(abs(np.array(maxA) - np.array(maxB)))
        maxSepSum[string] = np.nansum(_)
        maxSep[string] = _

    return (mlcData, jawData, segWeightData, maxSep, maxSepSum, maxPositions, maxJawPos)



# Copyright (C) 2019 Cancer Care Associates

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pymedphys._imports import numpy as np


def transform_axis(x, y, transform):
    transformed_x = transform @ np.vstack([x, np.zeros(len(x)), np.ones(len(x))])
    transformed_y = transform @ np.vstack([np.zeros(len(y)), y, np.ones(len(y))])

    return transformed_x[0:2], transformed_y[0:2]



import pathlib
import subprocess
import webbrowser

LIBRARY_ROOT = pathlib.Path(__file__).resolve().parent.parent
DOCS_DIR = LIBRARY_ROOT.joinpath("docs")
DOCS_BUILD_DIR = DOCS_DIR.joinpath("_build")
DOCS_HTML_BUILD_DIR = DOCS_BUILD_DIR.joinpath("html")


def build_docs(args):
    subprocess.check_call(
        "pandoc CHANGELOG.md --from markdown --to rst -s -o release-notes.rst",
        cwd=DOCS_DIR,
        shell=True,
    )

    if args.output:
        output_directory = args.output
    else:
        output_directory = str(DOCS_HTML_BUILD_DIR)

    if args.live:
        webbrowser.open("http://127.0.0.1:8000")

        subprocess.check_call(
            " ".join(["sphinx-autobuild", str(DOCS_DIR), output_directory]), shell=True
        )

    else:
        subprocess.check_call(
            " ".join(["sphinx-build", "-W", str(DOCS_DIR), output_directory]),
            shell=True,
        )



from pymedphys._imports import numpy as np

import matplotlib.pyplot as plt
from scipy import signal
from scipy.signal import find_peaks


# this subroutine aims to find the peaks input (resampled amplitude profile)
def peak_find(ampl_resamp, dx):
    peak_figs = []
    peaks = []
    peak_type = []
    for j in range(0, ampl_resamp.shape[1] - 1):
        amp_base_res = signal.convolve(
            ampl_resamp[:, j], ampl_resamp[:, j], mode="full"
        )
        amp_base_res = signal.resample(
            amp_base_res / np.amax(amp_base_res), int(np.ceil(len(amp_base_res) / 2))
        )
        for k in range(j + 1, ampl_resamp.shape[1]):
            amp_overlay_res = signal.convolve(
                ampl_resamp[:, k], ampl_resamp[:, k], mode="full"
            )
            amp_overlay_res = signal.resample(
                amp_overlay_res / np.amax(amp_overlay_res),
                int(np.ceil(len(amp_overlay_res) / 2)),
            )
            # amp_overlay_res = signal.savgol_filter(ampl_resamp[:, k], 1501, 1)

            peak1, _ = find_peaks(amp_base_res, prominence=0.5)
            peak2, _ = find_peaks(amp_overlay_res, prominence=0.5)

            if (
                abs(peak2 - peak1) < 2500
            ):  # if the two peaks are separated the two fields are not adjacent.
                amp_peak = ampl_resamp[:, j] + ampl_resamp[:, k]
                x = np.linspace(
                    0, 0 + (len(amp_peak) * dx / 10), len(amp_peak), endpoint=False
                )  # definition of the distance axis

                peak_pos, _ = find_peaks(
                    signal.savgol_filter(
                        amp_peak[min(peak1[0], peak2[0]) : max(peak1[0], peak2[0])],
                        201,
                        3,
                    ),
                    prominence=0.010,
                )
                pos_prominence = signal.peak_prominences(
                    signal.savgol_filter(
                        amp_peak[min(peak1[0], peak2[0]) : max(peak1[0], peak2[0])],
                        201,
                        3,
                    ),
                    peak_pos,
                )
                # print('#peaks pos det=', len(peak_pos), peak_pos)
                # print('#pos peaks prominence=', pos_prominence[0])
                peak_neg, _ = find_peaks(
                    signal.savgol_filter(
                        -amp_peak[min(peak1[0], peak2[0]) : max(peak1[0], peak2[0])],
                        201,
                        3,
                    ),
                    prominence=0.010,
                )
                neg_prominence = signal.peak_prominences(
                    signal.savgol_filter(
                        -amp_peak[min(peak1[0], peak2[0]) : max(peak1[0], peak2[0])],
                        201,
                        3,
                    ),
                    peak_neg,
                )
                # print('#peaks neg det=',len(peak_neg),peak_neg)
                # print('#neg peaks prominence=', neg_prominence[0])
                # we now need to select the peak with the largest prominence positve or negative
                # we add all the peaks and prominences toghether
                peaks_all = np.concatenate((peak_pos, peak_neg), axis=None)
                prom_all = np.concatenate(
                    (pos_prominence[0], neg_prominence[0]), axis=None
                )
                # print('all peaks',peaks_all,prom_all)

                if peaks_all.size != 0:
                    peak = peaks_all[np.argmax(prom_all)]
                    if peak in peak_pos:
                        peak_type.append(1)
                        peaks.append(min(peak1[0], peak2[0]) + peak)
                        # print('pos peak')
                    elif peak in peak_neg:
                        peak_type.append(0)
                        peaks.append(min(peak1[0], peak2[0]) + peak)
                        # print('neg peak')

                    fig = plt.figure(figsize=(10, 6))
                    plt.plot(x, amp_peak, label="Total amplitude profile")
                    plt.plot(
                        x[min(peak1[0], peak2[0]) + peak],
                        amp_peak[min(peak1[0], peak2[0]) + peak],
                        "x",
                        label="Peaks detected",
                    )
                    plt.ylabel("amplitude [a.u.]")
                    plt.xlabel("distance [mm]")
                    plt.legend()
                    fig.suptitle("Junctions", fontsize=16)
                    peak_figs.append(fig)

                elif peaks_all.size == 0:
                    peaks.append(0)
                    peak_type.append(0)
                    print("no peak has been found")
                    fig = plt.figure(figsize=(10, 6))
                    plt.plot(x, amp_peak, label="Total amplitude profile")
                    # plt.plot(x[min(peak1[0], peak2[0]) + peak], amp_peak[min(peak1[0], peak2[0]) + peak], "x",
                    #          label='Peaks detected')
                    plt.ylabel("amplitude [a.u.]")
                    plt.xlabel("distance [mm]")
                    plt.legend()
                    fig.suptitle("Junctions", fontsize=16)
                    peak_figs.append(fig)

            # else:
            # print(j, k, 'the data is not contiguous finding another curve in dataset')

    # print('peaks_here=',peaks)
    return peaks, peak_type, peak_figs


# this subroutine aims to find the peaks
def peak_find_fieldrot(ampl_resamp, dx, profilename):
    peaks = []
    peak_type = []
    for j in range(0, ampl_resamp.shape[1] - 1):
        amp_base_res = signal.convolve(
            ampl_resamp[:, j], ampl_resamp[:, j], mode="full"
        )
        amp_base_res = signal.resample(
            amp_base_res / np.amax(amp_base_res), int(np.ceil(len(amp_base_res) / 2))
        )

        for k in range(j + 1, ampl_resamp.shape[1]):
            amp_overlay_res = signal.convolve(
                ampl_resamp[:, k], ampl_resamp[:, k], mode="full"
            )
            amp_overlay_res = signal.resample(
                amp_overlay_res / np.amax(amp_overlay_res),
                int(np.ceil(len(amp_overlay_res) / 2)),
            )

            peak1, _ = find_peaks(amp_base_res, prominence=0.5)
            peak2, _ = find_peaks(amp_overlay_res, prominence=0.5)
            # print('peak find', peak1, peak2, abs(peak2 - peak1),len(amp_base_res))

            if (
                abs(peak2 - peak1) <= 4000
            ):  # if the two peaks are separated the two fields are not adjacent.
                amp_peak = ampl_resamp[:, j] + ampl_resamp[:, k]
                x = np.linspace(
                    0, 0 + (len(amp_peak) * dx / 10), len(amp_peak), endpoint=False
                )  # definition of the distance axis

                peak_pos, _ = find_peaks(
                    signal.savgol_filter(
                        amp_peak[min(peak1[0], peak2[0]) : max(peak1[0], peak2[0])],
                        201,
                        3,
                    ),
                    prominence=0.010,
                )
                pos_prominence = signal.peak_prominences(
                    signal.savgol_filter(
                        amp_peak[min(peak1[0], peak2[0]) : max(peak1[0], peak2[0])],
                        201,
                        3,
                    ),
                    peak_pos,
                )
                # print('#peaks pos det=', len(peak_pos), peak_pos)
                # print('#pos peaks prominence=', pos_prominence[0])
                peak_neg, _ = find_peaks(
                    signal.savgol_filter(
                        -amp_peak[min(peak1[0], peak2[0]) : max(peak1[0], peak2[0])],
                        201,
                        3,
                    ),
                    prominence=0.010,
                )
                neg_prominence = signal.peak_prominences(
                    signal.savgol_filter(
                        -amp_peak[min(peak1[0], peak2[0]) : max(peak1[0], peak2[0])],
                        201,
                        3,
                    ),
                    peak_neg,
                )
                # print('#peaks neg det=',len(peak_neg),peak_neg)
                # print('#neg peaks prominence=', neg_prominence[0])
                # we now need to select the peak with the largest prominence positve or negative
                # we add all the peaks and prominences toghether
                peaks_all = np.concatenate((peak_pos, peak_neg), axis=None)
                prom_all = np.concatenate(
                    (pos_prominence[0], neg_prominence[0]), axis=None
                )
                # print('all peaks',peaks_all,prom_all)
                peak = peaks_all[np.argmax(prom_all)]
                if peak.size != 0:
                    if peak in peak_pos:
                        peak_type.append(1)
                        peaks.append(min(peak1[0], peak2[0]) + peak)
                        # print('pos peak')
                    elif peak in peak_neg:
                        peak_type.append(0)
                        peaks.append(min(peak1[0], peak2[0]) + peak)
                        # print('neg peak')

                    fig = plt.figure(figsize=(10, 6))
                    plt.plot(
                        x, amp_peak, label="Total amplitude profile - " + profilename
                    )
                    plt.plot(
                        x[min(peak1[0], peak2[0]) + peak],
                        amp_peak[min(peak1[0], peak2[0]) + peak],
                        "x",
                        label="Peaks detected",
                    )
                    plt.ylabel("amplitude [a.u.]")
                    plt.xlabel("distance [mm]")
                    plt.legend()
                    fig.suptitle("Junctions - " + profilename, fontsize=16)

                elif peak.size == 0:
                    peaks.append(0)
                    peak_type.append(0)
                    print("no peak has been found")

            # else:
            # print(j, k, 'the data is not contiguous finding another curve in dataset')

    # print('peaks=',peaks)
    return peaks, peak_type, fig



# Copyright (C) 2020 University of New South Wales & Ingham Institute

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import signal
import sys
import tempfile
import pathlib
import uuid

from pymedphys._imports import pydicom, pynetdicom
from pymedphys._dicom.connect.base import DicomConnectBase
from pymedphys._dicom.constants.core import DICOM_SOP_CLASS_NAMES_MODE_PREFIXES


class DicomListener(DicomConnectBase):
    """Class which provides SCP functionality to listen for incoming DICOM objects
    """

    def __init__(self, storage_directory=None, on_released_callback=None, **kwargs):
        """Create and instance of a DICOM Listener

        Parameters
        ----------
        storage_directory : pathlib.Path or str, optional
            The directory in which to store incoming DICOM objects, by default a
            temporary directory will be created
        on_released_callback : function, optional
            Called when an association is released, the directory in which the incoming
            data was stored is returned, by default None
        """
        super().__init__(**kwargs)

        # If no storage directory is set, create a temporary directory
        if storage_directory is None:
            storage_directory = tempfile.mkdtemp()

        # The directory where incoming data will be stored
        self.storage_directory = pathlib.Path(storage_directory)

        # Callback when association is released called with path to the directory where
        # data from that association was stored
        self.on_released_callback = on_released_callback

        # Keep track of the directory where data for the current association is stored
        self.association_directory = None

        # The application entity
        self.ae = None

        logging.debug("Will store files received in: %s", self.storage_directory)

    def start(self):
        """Start the DICOM listener
        """

        # Initialise the Application Entity
        self.ae = pynetdicom.AE(ae_title=self.ae_title)

        # Add the supported presentation context
        self.ae.add_supported_context(
            pynetdicom.sop_class.VerificationSOPClass  # pylint: disable = no-member
        )
        for context in pynetdicom.StoragePresentationContexts:
            self.ae.add_supported_context(context.abstract_syntax)

        handlers = [
            (pynetdicom.evt.EVT_C_STORE, self.on_c_store),
            (pynetdicom.evt.EVT_C_ECHO, self.on_c_echo),
            (pynetdicom.evt.EVT_ACCEPTED, self.on_association_accepted),
            (pynetdicom.evt.EVT_RELEASED, self.on_association_released),
        ]

        # Start listening for incoming association requests
        self.ae.start_server((self.host, self.port), evt_handlers=handlers, block=False)

    def stop(self):
        """Stop the DICOM listener
        """

        if self.ae:
            self.ae.shutdown()

    def on_c_echo(self, _):  # pylint: disable = no-self-use
        """Respond to a C-ECHO service request.
        """
        logging.debug("C-ECHO!")
        return 0x0000

    def on_association_accepted(self, _):
        self.association_directory = None

    def on_association_released(self, _):
        if self.on_released_callback:
            self.on_released_callback(self.association_directory)

    def on_c_store(self, event):

        dataset = event.dataset

        try:
            mode_prefix = DICOM_SOP_CLASS_NAMES_MODE_PREFIXES[dataset.SOPClassUID.name]
        except KeyError:
            mode_prefix = "UN"

        series_uid = dataset.SeriesInstanceUID
        series_dir = self.storage_directory.joinpath(series_uid)
        series_dir.mkdir(exist_ok=True)
        self.association_directory = series_dir

        filename = "{0!s}.{1!s}".format(mode_prefix, dataset.SOPInstanceUID)
        filepath = series_dir.joinpath(filename)

        status_ds = pydicom.Dataset()
        status_ds.Status = 0x0000

        if filepath.exists():

            # If the file already exists, open it up and compare the hash of it's
            # contents (converting to JSON string to perform hash)
            existing_ds = pydicom.read_file(filepath)
            existing_hash = hash(existing_ds.to_json())
            incoming_hash = hash(dataset.to_json())

            if not existing_hash == incoming_hash:
                # If the hashes don't match, save conflicting incoming file in an
                # "orphan" sub-directory.

                # Just give the file a unique name in the orphan directory
                filename = str(uuid.uuid4())
                filepath = series_dir.joinpath("orphan", filename)
                filepath.parent.mkdir(exist_ok=True)

                logging.warning(
                    "DICOM file exists, storing in orphan directory: %s", filename
                )

        context = event.context
        meta = pydicom.Dataset()
        meta.MediaStorageSOPClassUID = dataset.SOPClassUID
        meta.MediaStorageSOPInstanceUID = dataset.SOPInstanceUID
        meta.ImplementationClassUID = pynetdicom.PYNETDICOM_IMPLEMENTATION_UID
        meta.TransferSyntaxUID = context.transfer_syntax

        # The following is not mandatory, set for convenience
        meta.ImplementationVersionName = pynetdicom.PYNETDICOM_IMPLEMENTATION_VERSION
        file_ds = pydicom.FileDataset(
            filepath, {}, file_meta=meta, preamble=b"\0" * 128
        )
        file_ds.update(dataset)
        file_ds.is_little_endian = context.transfer_syntax.is_little_endian
        file_ds.is_implicit_VR = context.transfer_syntax.is_implicit_VR

        try:
            # We use `write_like_original=False` to ensure that a compliant
            # File Meta Information Header is written
            file_ds.save_as(filepath, write_like_original=False)
            status_ds.Status = 0x0000  # Success
        except IOError:
            logging.error("Could not write file to specified directory:")
            logging.error("    %s", filepath)
            logging.error(
                "Directory may not exist or you may not have write permission"
            )

            status_ds.ErrorComment = "SCP internal error - Unable to write file"
            status_ds.Status = 0xA700  # Failed - Out of Resources - IOError

        return status_ds


def listen_cli(args):
    """Start a DICOM listener from the command line interface
    """

    # Set log level to debug
    root = logging.getLogger()
    root.setLevel(logging.DEBUG)

    # Start the listener
    dicom_listener = DicomListener(
        port=args.port, ae_title=args.aetitle, storage_directory=args.storage_directory
    )
    dicom_listener.start()

    # Run until the process is stopped
    def handler_stop_signals(*_):
        logging.info("Shutting down listener")
        dicom_listener.stop()
        sys.exit()

    signal.signal(signal.SIGINT, handler_stop_signals)
    signal.signal(signal.SIGTERM, handler_stop_signals)

    while True:
        pass



# Copyright (C) 2020 Simon Biggs

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pymedphys._experimental.pseudonymisation import anonymise_with_pseudo_cli
from pymedphys.cli import dicom


def dicom_cli(subparsers):
    dicom_parser, dicom_subparsers = dicom.set_up_dicom_cli(subparsers)
    anonymise(dicom_subparsers)

    return dicom_parser


def anonymise(dicom_subparsers):
    parser = dicom.anonymise(dicom_subparsers)

    parser.add_argument(
        "--pseudo",
        action="store_true",
        help=(
            "Use this flag to activate the use of pseudonymisation of the "
            "identifying elements in the DICOM files, "
            "as opposed to replacing them with 'dummy' values. "
            "The pseudonymised values are SHA3_256 hashed if text or UIDs, shifted if dates, and jittered if an age. "
            "UIDs retain consistency so that entire sets of data will retain their referential integrity. "
            "Patient Identifiers stay consistent but are not reversible, and are unlikely to collide with other patient identifiers"
        ),
    )

    parser.set_defaults(func=anonymise_with_pseudo_cli)



# Copyright (C) 2019 Paul King, Simon Biggs
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# The following needs to be removed before leaving the experimental module
# pylint: skip-file


""" Test profile. """

import os

import numpy as np

from pymedphys._data import download
from pymedphys._experimental.paulking.profile import Profile

PROFILER = [
    (-16.4, 0.22),
    (-16, 0.3),
    (-15.6, 0.28),
    (-15.2, 0.3),
    (-14.8, 0.36),
    (-14.4, 0.38),
    (-14, 0.41),
    (-13.6, 0.45),
    (-13.2, 0.47),
    (-12.8, 0.51),
    (-12.4, 0.55),
    (-12, 0.62),
    (-11.6, 0.67),
    (-11.2, 0.74),
    (-10.8, 0.81),
    (-10.4, 0.91),
    (-10, 0.97),
    (-9.6, 1.12),
    (-9.2, 1.24),
    (-8.8, 1.4),
    (-8.4, 1.56),
    (-8, 1.75),
    (-7.6, 2.06),
    (-7.2, 2.31),
    (-6.8, 2.56),
    (-6.4, 3.14),
    (-6, 3.83),
    (-5.6, 4.98),
    (-5.2, 8.17),
    (-4.8, 40.6),
    (-4.4, 43.34),
    (-4, 44.17),
    (-3.6, 44.44),
    (-3.2, 44.96),
    (-2.8, 44.18),
    (-2.4, 45.16),
    (-2, 45.54),
    (-1.6, 45.07),
    (-1.2, 45.28),
    (-0.8, 45.27),
    (-0.4, 44.57),
    (0, 45.23),
    (0.4, 45.19),
    (0.8, 45.18),
    (1.2, 45.37),
    (1.6, 45.34),
    (2, 45.39),
    (2.4, 45.32),
    (2.8, 45.25),
    (3.2, 44.84),
    (3.6, 44.76),
    (4, 44.23),
    (4.4, 43.22),
    (4.8, 39.14),
    (5.2, 7.98),
    (5.6, 4.89),
    (6, 3.71),
    (6.4, 3.11),
    (6.8, 2.59),
    (7.2, 2.27),
    (7.6, 1.95),
    (8, 1.71),
    (8.4, 1.46),
    (8.8, 1.35),
    (9.2, 1.18),
    (9.6, 1.11),
    (10, 0.93),
    (10.4, 0.87),
    (10.8, 0.78),
    (11.2, 0.7),
    (11.6, 0.64),
    (12, 0.6),
    (12.4, 0.54),
    (12.8, 0.49),
    (13.2, 0.47),
    (13.6, 0.43),
    (14, 0.4),
    (14.4, 0.39),
    (14.8, 0.34),
    (15.2, 0.33),
    (15.6, 0.32),
    (16, 0.3),
    (16.4, 0.3),
]

WEDGED = [
    (-16.4, 0.27),
    (-16, 0.31),
    (-15.6, 0.29),
    (-15.2, 0.29),
    (-14.8, 0.32),
    (-14.4, 0.33),
    (-14, 0.35),
    (-13.6, 0.38),
    (-13.2, 0.4),
    (-12.8, 0.44),
    (-12.4, 0.46),
    (-12, 0.51),
    (-11.6, 0.55),
    (-11.2, 0.6),
    (-10.8, 0.65),
    (-10.4, 0.7),
    (-10, 0.74),
    (-9.6, 0.84),
    (-9.2, 0.94),
    (-8.8, 1.04),
    (-8.4, 1.14),
    (-8, 1.25),
    (-7.6, 1.45),
    (-7.2, 1.6),
    (-6.8, 1.78),
    (-6.4, 2.14),
    (-6, 2.66),
    (-5.6, 3.62),
    (-5.2, 6.54),
    (-4.8, 17.55),
    (-4.4, 20.07),
    (-4, 21.37),
    (-3.6, 22.19),
    (-3.2, 23.1),
    (-2.8, 23.74),
    (-2.4, 24.56),
    (-2, 25.49),
    (-1.6, 26.35),
    (-1.2, 27),
    (-0.8, 28.06),
    (-0.4, 28.89),
    (0, 29.8),
    (0.4, 30.61),
    (0.8, 31.4),
    (1.2, 32.53),
    (1.6, 33.06),
    (2, 34.15),
    (2.4, 34.85),
    (2.8, 35.65),
    (3.2, 36.6),
    (3.6, 37.04),
    (4, 37.45),
    (4.4, 36.72),
    (4.8, 30.93),
    (5.2, 10.06),
    (5.6, 5.43),
    (6, 3.71),
    (6.4, 3.01),
    (6.8, 2.52),
    (7.2, 2.19),
    (7.6, 1.9),
    (8, 1.7),
    (8.4, 1.48),
    (8.8, 1.35),
    (9.2, 1.19),
    (9.6, 1.09),
    (10, 0.93),
    (10.4, 0.89),
    (10.8, 0.78),
    (11.2, 0.72),
    (11.6, 0.65),
    (12, 0.6),
    (12.4, 0.55),
    (12.8, 0.5),
    (13.2, 0.48),
    (13.6, 0.45),
    (14, 0.41),
    (14.4, 0.4),
    (14.8, 0.35),
    (15.2, 0.33),
    (15.6, 0.32),
    (16, 0.31),
    (16.4, 0.3),
]


def get_data_file(filename):
    return download.get_file_within_data_zip("paulking_test_data.zip", filename)


def test_init():
    assert np.allclose(Profile(x=[0], y=[0]).x, [0])


def test_interp():
    assert Profile().interp is None
    assert np.isclose(Profile(x=[0, 1], y=[0, 1]).interp(0.5), 0.5)


def test_magic_methods():
    assert not Profile()
    # __len__
    assert len(Profile().from_tuples(PROFILER)) == 83
    # __eq__
    assert Profile() == Profile()
    assert Profile(x=[], y=[]) == Profile()
    assert Profile(x=[0], y=[0]) != Profile()
    # __copy__
    original = Profile()
    same = original
    assert same == original
    # __str__
    empty_profile = Profile()
    print(empty_profile)
    profiler = Profile().from_tuples(PROFILER)
    assert profiler.__str__()
    # __add__, __radd__, __iadd__
    profiler = Profile().from_tuples(PROFILER)
    assert np.isclose(profiler.get_y(0), (profiler + 2).get_y(2))
    # __sub__, __rsub__, __isub__
    profiler = Profile().from_tuples(PROFILER)
    assert np.isclose(profiler.get_y(0), (profiler - 2).get_y(-2))
    # __mul__, __rmul__, __imul__
    profiler = Profile().from_tuples(PROFILER)
    assert np.isclose(4 * sum(profiler.y), sum((4 * profiler).y))
    assert np.isclose(4 * sum(profiler.y), sum((profiler * 4).y))
    ref = 4 * sum(profiler.y)
    profiler *= 4
    assert np.isclose(sum(profiler.y), ref)


def test_from_lists():
    empty = Profile()
    also_empty = empty
    also_empty.from_lists([], [])
    assert empty == also_empty


def test_from_tuples():
    empty = Profile()
    profiler = empty.from_tuples(PROFILER)
    assert len(profiler.x) == len(PROFILER)
    assert profiler.x[0] == PROFILER[0][0]


def test_from_pulse():
    pulse = 4 * Profile().from_pulse(0.0, 1, (-5, 5), 0.1)
    assert np.isclose(sum(pulse.y), 40)


def test_from_snc_profiler():
    file_name = get_data_file("test_varian_open.prs")

    x_profile = Profile().from_snc_profiler(file_name, "tvs")
    y_profile = Profile().from_snc_profiler(file_name, "rad")
    assert np.isclose(x_profile.get_y(0), 45.50562901780488)
    assert np.isclose(y_profile.get_y(0), 45.50562901780488)
    assert x_profile.meta["SSD"] == y_profile.meta["SSD"]


def test_from_narrow_png():
    file_name = get_data_file("FilmCalib_EBT_vert_strip.png")

    png = Profile().from_narrow_png(file_name)
    assert np.isclose(png.get_y(0), 0.609074819347117)


def test_get_y():
    profiler = Profile().from_tuples(PROFILER)
    assert np.isclose(profiler.get_y(0), 45.23)


def test_get_x():
    profiler = Profile().from_tuples(PROFILER)
    assert np.allclose(profiler.get_x(10), (-5.17742830712, 5.1740693196))


def test_get_increment():
    profiler = Profile().from_tuples(PROFILER)
    assert np.isclose(profiler.get_increment(), 0.4)


def test_slice_segment():
    profiler = Profile().from_tuples(PROFILER)
    # NO POINTS
    no_points = profiler.slice_segment(start=1, stop=0)
    assert np.array_equal(no_points.x, [])
    assert np.array_equal(no_points.y, [])
    # ONE POINT
    profiler = Profile().from_tuples(PROFILER)
    one_point = profiler.slice_segment(start=0, stop=0)
    assert np.array_equal(one_point.x, [0])
    assert np.array_equal(one_point.y, [45.23])
    # ALL POINTS
    profiler = Profile().from_tuples(PROFILER)
    all_points = profiler.slice_segment()
    assert np.array_equal(all_points.x, profiler.x)
    assert np.array_equal(all_points.y, profiler.y)


def test_resample_x():
    profiler = Profile().from_tuples(PROFILER, meta={"depth": 10})
    assert profiler.meta["depth"] == 10
    assert np.isclose(profiler.interp(0), profiler.resample_x(0.1).interp(0))
    assert np.isclose(profiler.interp(6.372), profiler.resample_x(0.1).interp(6.372))
    resampled = profiler.resample_x(0.1)
    increments = np.diff([i for i in resampled.x])
    assert np.allclose(increments, 0.1)
    assert np.isclose(resampled.y[0], profiler.y[0])


def test_resample_y():
    profiler = Profile().from_tuples(PROFILER)
    assert len(profiler.resample_y(0.5)) > len(profiler.resample_y(1))


def test_make_normal_y():
    profiler = Profile().from_tuples(PROFILER)
    assert np.isclose(profiler.make_normal_y(x=0).get_y(0), 1.0)


def test_get_edges():
    profiler = Profile().from_tuples(PROFILER)
    assert np.allclose(profiler.get_edges(), (-5.2, 4.8))
    assert len(profiler) == len(PROFILER)


def test_make_normal_x():
    profiler = Profile().from_tuples(PROFILER)
    assert np.isclose(profiler.make_normal_x().x[0], -3.1538461538461533)
    profiler = Profile().from_tuples(PROFILER)
    assert len(PROFILER) == len(profiler.make_normal_x().x)


def test_slice_umbra():
    profiler = Profile().from_tuples(PROFILER).resample_x(0.1)
    profiler_length = len(profiler)
    umbra = profiler.slice_umbra()
    assert len(umbra) < profiler_length


def test_slice_penumbra():
    profiler = Profile().from_tuples(PROFILER).resample_x(0.1)
    lt_penum, rt_penum = profiler.slice_penumbra()
    assert np.all(lt_penum.x < 0)
    assert np.all(rt_penum.x > 0)
    assert np.all(lt_penum.y < profiler.get_y(0))
    assert np.all(rt_penum.y < profiler.get_y(0))


def test_slice_shoulders():
    profiler = Profile().from_tuples(PROFILER).resample_x(0.1)
    lt_should, rt_should = profiler.slice_shoulders()
    assert np.all(lt_should.x < min(rt_should.x))
    assert np.all(rt_should.x > max(lt_should.x))


def test_slice_tails():
    profiler = Profile().from_tuples(PROFILER).resample_x(0.1)
    lt_tail, rt_tail = profiler.slice_tails()
    assert np.all(lt_tail.x < min(rt_tail.x))
    assert np.all(rt_tail.x > max(lt_tail.x))


def test_get_flatness():
    profiler = Profile().from_tuples(PROFILER)
    profiler = profiler.resample_x(0.1)
    assert np.isclose(profiler.get_flatness(), 0.03042644213284108)


def test_get_symmetry():
    profiler = Profile().from_tuples(PROFILER)
    profiler = profiler.resample_x(0.1)
    symmetry = profiler.get_symmetry()
    assert np.isclose(symmetry, 0.024152376510553037)


def test_make_symmetric():
    profiler = Profile().from_tuples(PROFILER)
    assert np.isclose(profiler.make_symmetric().get_symmetry(), 0.0)


def test_make_centered():
    profiler = Profile().from_tuples(PROFILER)
    assert np.isclose(np.sum(profiler.make_centered().get_edges()), 0.0)


def test_make_flipped():
    profiler = Profile().from_tuples(PROFILER)
    assert np.isclose(profiler.get_y(3), profiler.make_flipped().get_y(-3))


def test_align_to():
    profiler = Profile().from_tuples(PROFILER)
    assert np.isclose(profiler.align_to(profiler + (2)).x[0], profiler.x[0] + 2)


def test_cross_calibrate():
    reference_file_name = get_data_file("FilmCalib.prs")
    measured_file_name = get_data_file("FilmCalib_EBT_vert_strip.png")

    cal_curve = Profile().cross_calibrate(reference_file_name, measured_file_name)
    assert min(cal_curve.x) <= 1
    assert max(cal_curve.x) >= 0



# Copyright (C) 2020 Cancer Care Associates

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pymedphys._imports import pydicom
from pymedphys._imports import streamlit as st

import pymedphys
from pymedphys._dicom.constants.uuid import DICOM_PLAN_UID
from pymedphys._streamlit.apps.metersetmap import _config
from pymedphys._streamlit.utilities import exceptions as _exceptions
from pymedphys._streamlit.utilities import misc as st_misc
from pymedphys._utilities import patient as utl_patient


def pydicom_hash_function(dicom):
    return hash(dicom.SOPInstanceUID)


@st.cache(hash_funcs={pydicom.dataset.FileDataset: pydicom_hash_function})
def load_dicom_file_if_plan(filepath):
    dcm = pydicom.read_file(str(filepath), force=True, stop_before_pixels=True)
    if dcm.SOPClassUID == DICOM_PLAN_UID:
        return dcm

    return None


def dicom_input_method(  # pylint: disable = too-many-return-statements
    key_namespace="", patient_id="", site=None, **_
):
    monaco_site = site

    FILE_UPLOAD = "File upload"
    MONACO_SEARCH = "Search Monaco file export location"

    import_method = st.radio(
        "DICOM import method",
        [FILE_UPLOAD, MONACO_SEARCH],
        key=f"{key_namespace}_dicom_file_import_method",
    )

    if import_method == FILE_UPLOAD:
        dicom_plan_bytes = st.file_uploader(
            "Upload DICOM RT Plan File", key=f"{key_namespace}_dicom_plan_uploader"
        )

        if dicom_plan_bytes is None:
            return {}

        try:
            dicom_plan_bytes.seek(0)
            dicom_plan = pydicom.read_file(dicom_plan_bytes, force=True)
        except:  # pylint: disable = bare-except
            st.write(_exceptions.WrongFileType("Does not appear to be a DICOM file"))
            return {}

        if dicom_plan.SOPClassUID != DICOM_PLAN_UID:
            st.write(
                _exceptions.WrongFileType(
                    "The DICOM type needs to be an RT DICOM Plan file"
                )
            )
            return {}

        data_paths = []

    if import_method == MONACO_SEARCH:
        try:
            dicom_export_locations = _config.get_dicom_export_locations()
        except KeyError:
            st.write(
                _exceptions.ConfigMissing(
                    "No Monaco directory is configured. Please use "
                    f"'{FILE_UPLOAD}' instead."
                )
            )
            return {}

        monaco_site = st_misc.site_picker(
            "Monaco Export Location",
            default=monaco_site,
            key=f"{key_namespace}_monaco_site",
        )

        monaco_export_directory = dicom_export_locations[monaco_site]
        st.write(monaco_export_directory.resolve())

        patient_id = st.text_input(
            "Patient ID", patient_id, key=f"{key_namespace}_patient_id"
        )

        found_dicom_files = list(monaco_export_directory.glob(f"{patient_id}_*.dcm"))

        dicom_plans = {}

        for path in found_dicom_files:
            dcm = load_dicom_file_if_plan(path)
            if dcm is not None:
                dicom_plans[path.name] = dcm

        dicom_plan_options = list(dicom_plans.keys())

        if len(dicom_plan_options) == 0 and patient_id != "":
            st.write(
                _exceptions.NoRecordsFound(
                    f"No exported DICOM RT plans found for Patient ID {patient_id} "
                    f"within the directory {monaco_export_directory}"
                )
            )
            return {"patient_id": patient_id}

        if len(dicom_plan_options) == 1:
            selected_plan = dicom_plan_options[0]
        else:
            selected_plan = st.radio(
                "Select DICOM Plan",
                dicom_plan_options,
                key=f"{key_namespace}_select_monaco_export_plan",
            )

        st.write(f"DICOM file being used: `{selected_plan}`")

        dicom_plan = dicom_plans[selected_plan]
        data_paths = [monaco_export_directory.joinpath(selected_plan)]

    patient_id = str(dicom_plan.PatientID)
    st.write(f"Patient ID: `{patient_id}`")

    patient_name = str(dicom_plan.PatientName)
    patient_name = utl_patient.convert_patient_name(patient_name)

    st.write(f"Patient Name: `{patient_name}`")

    rt_plan_name = str(dicom_plan.RTPlanName)
    st.write(f"Plan Name: `{rt_plan_name}`")

    try:
        deliveries_all_fractions = pymedphys.Delivery.from_dicom(
            dicom_plan, fraction_number="all"
        )
    except AttributeError:
        st.write(_exceptions.WrongFileType("Does not appear to be a photon DICOM plan"))
        return {}

    fractions = list(deliveries_all_fractions.keys())
    if len(fractions) == 1:
        delivery = deliveries_all_fractions[fractions[0]]
    else:
        fraction_choices = {}

        for fraction, delivery in deliveries_all_fractions.items():
            rounded_mu = round(delivery.mu[-1], 1)

            fraction_choices[f"Perscription {fraction} with {rounded_mu} MU"] = fraction

        fraction_selection = st.radio(
            "Select relevant perscription",
            list(fraction_choices.keys()),
            key=f"{key_namespace}_dicom_perscription_chooser",
        )

        fraction_number = fraction_choices[fraction_selection]
        delivery = deliveries_all_fractions[fraction_number]

    deliveries = [delivery]

    identifier = f"DICOM ({rt_plan_name})"

    return {
        "site": monaco_site,
        "patient_id": patient_id,
        "patient_name": patient_name,
        "data_paths": data_paths,
        "identifier": identifier,
        "deliveries": deliveries,
    }



import pathlib
import time

from pymedphys._imports import watchdog

import pymedphys._utilities.filesystem


def read_and_trigger_callback(event, callback):
    event_source_path = pathlib.Path(event.src_path)
    ip = event_source_path.parent.name

    with pymedphys._utilities.filesystem.open_no_lock(  # pylint: disable = protected-access
        event_source_path, "rb"
    ) as f:
        data = f.read()

    callback(ip, data)


def create_event_handler(callback):
    def on_created(event):
        print(f"File created: {event.src_path}")
        read_and_trigger_callback(event, callback)

    def on_deleted(_):
        pass

    def on_modified(event):
        print(f"File modified: {event.src_path}")
        read_and_trigger_callback(event, callback)

    def on_moved(_):
        pass

    event_handler = watchdog.events.PatternMatchingEventHandler(
        patterns="*/[0-2][0-9][0-9].txt",
        ignore_patterns="",
        ignore_directories=True,
        case_sensitive=True,
    )
    event_handler.on_created = on_created
    event_handler.on_deleted = on_deleted
    event_handler.on_modified = on_modified
    event_handler.on_moved = on_moved

    return event_handler


def observe_with_callback(directories_to_watch, callback):
    event_handler = create_event_handler(callback)

    observers = []

    for watch_path in directories_to_watch:
        observer = watchdog.observers.Observer()
        # observer = watchdog.observers.polling.PollingObserver()
        observer.schedule(event_handler, watch_path, recursive=True)

        observers.append(observer)

    for observer in observers:
        observer.start()

    try:
        while True:
            time.sleep(1)
    finally:
        for observer in observers:
            observer.stop()
            observer.join()



# Copyright (C) 2019 Simon Biggs

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import hashlib
import json
import pathlib
import urllib

from pymedphys._imports import keyring, requests

from .zenodo import get_zenodo_access_token, get_zenodo_record_id

HEADERS = {"Content-Type": "application/json"}


def zenodo_api_with_helpful_fallback(url, method, **kwargs):
    hostname = urllib.parse.urlparse(url).hostname

    access_token = get_zenodo_access_token(hostname)
    kwargs["params"] = {"access_token": access_token}

    r = getattr(requests, method)(url, **kwargs)
    if r.status_code == 401:
        print("The access token you provided is invalid.\n")
        keyring.delete_password("Zenodo", hostname)
        return zenodo_api_with_helpful_fallback(url, **kwargs)
    if r.status_code == 403:
        print(
            "The access token you provided doesn't appear to have the right scopes. "
            "Make sure that the access token you provide has the scopes "
            "`deposit:actions`, `deposit:write`, and `user:email`.\n"
        )
        keyring.delete_password("Zenodo", hostname)
        return zenodo_api_with_helpful_fallback(url, **kwargs)

    return r


def create_metadata(title, author=None):
    if author is None:
        author = "PyMedPhys Contributors"

    metadata = {
        "metadata": {
            "title": title,
            "upload_type": "dataset",
            "creators": [{"name": author}],
            "description": "<p>This is an automated upload from the PyMedPhys library.</p>",
            "license": "Apache-2.0",
            "access_right": "open",
        }
    }

    return json.dumps(metadata)


def publish_deposition(deposition_id, use_sandbox=False):
    deposition_url = get_deposition_url(deposition_id, use_sandbox=use_sandbox)

    publish_url = f"{deposition_url}/actions/publish"
    r = zenodo_api_with_helpful_fallback(publish_url, "post")

    if r.status_code != 202:
        raise ValueError(
            f"Unexpected status code when publishing the file. Expected 202, got {r.status_code}."
        )


def get_zenodo_url(use_sandbox):
    if use_sandbox:
        zenodo_url = "https://sandbox.zenodo.org/"
    else:
        zenodo_url = "https://zenodo.org/"

    return zenodo_url


def get_deposition_url(deposition_id, use_sandbox=False):
    root_depositions_url = get_root_depositions_url(use_sandbox)
    deposition_url = f"{root_depositions_url}/{deposition_id}"

    return deposition_url


def get_root_depositions_url(use_sandbox):
    zenodo_url = get_zenodo_url(use_sandbox)
    root_depositions_url = f"{zenodo_url}api/deposit/depositions"

    return root_depositions_url


def get_files_url(deposition_id, use_sandbox=False):
    deposition_url = get_deposition_url(deposition_id, use_sandbox=use_sandbox)
    files_url = f"{deposition_url}/files"

    return files_url


def delete_filenames_within_record(filenames, deposition_id, use_sandbox=False):
    files_url = get_files_url(deposition_id, use_sandbox=use_sandbox)

    r = zenodo_api_with_helpful_fallback(files_url, "get")
    record_filenames = [record["filename"] for record in r.json()]

    for filename in filenames:
        if filename in record_filenames:
            file_self_urls = [
                record["links"]["self"]
                for record in r.json()
                if record["filename"] == filename
            ]

            if len(file_self_urls) > 1:
                raise ValueError("Unexpected number of file_ids found")

            file_self_url = file_self_urls[0]

            zenodo_api_with_helpful_fallback(file_self_url, "delete")


def upload_filepaths(filepaths, deposition_id, use_sandbox=False):
    files_url = get_files_url(deposition_id, use_sandbox=use_sandbox)

    for filepath in filepaths:
        md5 = hashlib.md5()

        with open(filepath, "rb") as upload_file:
            md5.update(upload_file.read())

            upload_file.seek(0)
            r = zenodo_api_with_helpful_fallback(
                files_url,
                "post",
                data={"name": filepath.name},
                files={"file": upload_file},
            )

        response = r.json()

        try:
            record_checksum = response["checksum"]
        except KeyError:
            raise ValueError(
                f"The response when uploading was not as expected:\n {response}"
            )

        if md5.hexdigest() != record_checksum:
            raise ValueError(
                "The uploaded Zenodo's checksum does not match the local checksum.\n"
                f"  Zenodo's Checksum: {record_checksum}\n"
                f"  Local Checksum: {md5.hexdigest()}"
            )


def set_metadata(metadata, deposition_id, use_sandbox=False):
    deposition_url = get_deposition_url(deposition_id, use_sandbox=use_sandbox)
    zenodo_api_with_helpful_fallback(
        deposition_url, "put", data=metadata, headers=HEADERS
    )


def upload_files_to_zenodo(
    filepaths, title, author=None, use_sandbox=False, record_name=None
):
    filepaths = [pathlib.Path(filepath) for filepath in filepaths]

    root_depositions_url = get_root_depositions_url(use_sandbox)

    if record_name is not None:
        if use_sandbox:
            raise ValueError("Cannot use sandbox when `record_name` is provided")

        old_deposition_id = get_zenodo_record_id(record_name)
        old_deposition_url = f"{root_depositions_url}/{old_deposition_id}"
        new_version_url = f"{old_deposition_url}/actions/newversion"
        r = zenodo_api_with_helpful_fallback(new_version_url, "post")
        deposition_id = int(r.json()["links"]["latest_draft"].split("/")[-1])
    else:
        r = zenodo_api_with_helpful_fallback(
            root_depositions_url, "post", json={}, headers=HEADERS
        )

        try:
            response = r.json()
        except json.JSONDecodeError:
            raise ValueError(f"Unexpected response: {r.text}")

        try:
            deposition_id = response["id"]
        except KeyError:
            raise ValueError(f"Unexpected response: {response}")

    metadata = create_metadata(title, author)
    set_metadata(metadata, deposition_id, use_sandbox=use_sandbox)

    filenames = [path.name for path in filepaths]
    delete_filenames_within_record(filenames, deposition_id, use_sandbox=use_sandbox)

    upload_filepaths(filepaths, deposition_id, use_sandbox=use_sandbox)
    publish_deposition(deposition_id, use_sandbox=use_sandbox)

    return deposition_id
